# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Open-High-Low-Close binning."""

from typing import NamedTuple

import haiku as hk
import jax.numpy as jnp


class OHLCData(NamedTuple):
    OPEN: jnp.ndarray
    HIGH: jnp.ndarray
    LOW: jnp.ndarray
    CLOSE: jnp.ndarray


class OHLC(hk.Module):
    """Open-High-Low-Close binning."""

    def __init__(self, name=None):
        super().__init__(name=name)

    def __call__(self, input: jnp.ndarray, *, reset_on: bool):
        def init_array(shape, dtype):
            return jnp.full(shape, fill_value=jnp.nan, dtype=dtype)

        # states
        BIN_OPEN = hk.get_state("BIN_OPEN", input.shape, input.dtype, init_array)
        BIN_HIGH = hk.get_state("BIN_HIGH", input.shape, input.dtype, init_array)
        BIN_LOW = hk.get_state("BIN_LOW", input.shape, input.dtype, init_array)
        BIN_CLOSE = hk.get_state("BIN_CLOSE", input.shape, input.dtype, init_array)

        def daily_call(operand):
            input, BIN_CLOSE = operand
            # initialize OPEN, HIGH, LOW, CLOSE
            BIN_CLOSE = init_array(input.shape, input.dtype)
            BIN_OPEN = init_array(input.shape, input.dtype)
            BIN_HIGH = init_array(input.shape, input.dtype)
            BIN_LOW = init_array(input.shape, input.dtype)
            return OHLCData(BIN_OPEN, BIN_HIGH, BIN_LOW, BIN_CLOSE)

        operand = input, BIN_CLOSE
        BIN_state = OHLCData(BIN_OPEN, BIN_HIGH, BIN_LOW, BIN_CLOSE)
        updated_BIN_state = hk.cond(
            pred=reset_on,
            true_operand=operand,
            true_fun=daily_call,
            false_operand=operand,
            false_fun=(lambda operand: BIN_state),
        )
        BIN_OPEN, BIN_HIGH, BIN_LOW, BIN_CLOSE = updated_BIN_state

        # update BIN_CLOSE
        BIN_CLOSE = jnp.where(jnp.logical_not(jnp.isnan(input)), input, BIN_CLOSE)

        # initialize BIN_{OPEN,HIGH,LOW} with the first non nan BIN_CLOSE
        BIN_OPEN = jnp.where(
            jnp.isnan(BIN_OPEN),
            BIN_CLOSE,
            BIN_OPEN,
        )

        BIN_HIGH = jnp.where(
            jnp.isnan(BIN_HIGH),
            BIN_CLOSE,
            BIN_HIGH,
        )

        BIN_LOW = jnp.where(
            jnp.isnan(BIN_LOW),
            BIN_CLOSE,
            BIN_LOW,
        )

        # update BIN_HIGH and BIN_LOW
        BIN_HIGH = jnp.where(
            BIN_CLOSE > BIN_HIGH,
            BIN_CLOSE,
            BIN_HIGH,
        )
        BIN_LOW = jnp.where(
            BIN_CLOSE < BIN_LOW,
            BIN_CLOSE,
            BIN_LOW,
        )
        # update state
        hk.set_state("BIN_OPEN", BIN_OPEN)
        hk.set_state("BIN_HIGH", BIN_HIGH)
        hk.set_state("BIN_LOW", BIN_LOW)
        hk.set_state("BIN_CLOSE", BIN_CLOSE)

        return OHLCData(BIN_OPEN, BIN_HIGH, BIN_LOW, BIN_CLOSE)



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import overload

from typing_extensions import Literal

from . import modules
from .tensor import JAXTensor, NumPyTensor, PyTorchTensor, Tensor, TensorFlowTensor


@overload
def get_dummy(framework: Literal["pytorch"]) -> PyTorchTensor:
    ...


@overload
def get_dummy(framework: Literal["tensorflow"]) -> TensorFlowTensor:
    ...


@overload
def get_dummy(framework: Literal["jax"]) -> JAXTensor:
    ...


@overload
def get_dummy(framework: Literal["numpy"]) -> NumPyTensor:
    ...


@overload
def get_dummy(framework: str) -> Tensor:
    ...


def get_dummy(framework: str) -> Tensor:
    x: Tensor
    if framework == "pytorch":
        x = modules.torch.zeros(0)
        assert isinstance(x, PyTorchTensor)
    elif framework == "pytorch-gpu":
        x = modules.torch.zeros(0, device="cuda:0")  # pragma: no cover
        assert isinstance(x, PyTorchTensor)  # pragma: no cover
    elif framework == "tensorflow":
        x = modules.tensorflow.zeros(0)
        assert isinstance(x, TensorFlowTensor)
    elif framework == "jax":
        x = modules.jax.numpy.zeros(0)
        assert isinstance(x, JAXTensor)
    elif framework == "numpy":
        x = modules.numpy.zeros(0)
        assert isinstance(x, NumPyTensor)
    else:
        raise ValueError(f"unknown framework: {framework}")  # pragma: no cover
    return x.float32()



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from functools import partial, reduce

import jax.numpy as jnp
import numpy as onp
import pandas as pd
import xarray as xr
from haiku import transform_with_state
from jax import tree_flatten
from jax.numpy import array as DeviceArray
from jax.numpy import int32
from jax.tree_util import tree_map
from numpy import array
from pandas import NaT

import wax.external.eagerpy as ep
from wax.datasets.generate_temperature_data import (
    generate_temperature_data_multi_time_scale,
)
from wax.encode import encode_dataset
from wax.stream import (
    Stream,
    dataset_to_numpy,
    get_time_dataset,
    split_dataset_from_time_dims,
    tree_access_data,
    unroll_stream,
)
from wax.testing import assert_tree_all_close
from wax.unroll import unroll
from wax.utils import dict_map


def onp_half_precision(x):
    if x.dtype == onp.float64:
        return x.astype(onp.float32)
    if x.dtype == onp.int64:
        return x.astype(onp.int32)
    return x


def test_onp_half_precision():
    x = onp.array(1, dtype=onp.int32)
    assert onp_half_precision(x) == x


def generate_news_data():
    days = 2
    bins = 1000
    onp.random.seed(3)
    dates = pd.date_range(start="2000/01/02", periods=days, freq="B")
    news_times = pd.Index(
        reduce(
            lambda x, y: x + y,
            [
                (date + pd.timedelta_range(start=1, freq="1min", periods=bins)).tolist()
                for date in dates
            ],
        )
    )
    idx_news = onp.random.choice(len(news_times), 10)
    news_times = news_times[sorted(idx_news)]

    # create a bin with 5 observations
    news_times = news_times.tolist()
    news_times.append(news_times[0] + onp.timedelta64(3, "s"))
    news_times.append(news_times[0] + onp.timedelta64(10, "s"))
    news_times.append(news_times[0] + onp.timedelta64(20, "s"))
    news_times.append(news_times[0] + onp.timedelta64(27, "s"))
    news_times = pd.Index(sorted(news_times)).values
    news_times = news_times + onp.timedelta64(50, "ms")
    T = len(news_times)
    news_data = onp.random.normal(size=(T, 2)).astype(str)
    news_data = xr.DataArray(
        news_data,
        dims=("news_time", "news_attr"),
        coords={"news_time": news_times, "news_attr": ["head", "text"]},
    )
    return news_data


def prepare_test_data():
    xdata = generate_temperature_data_multi_time_scale()

    xdata["NEWS"] = generate_news_data()

    xdata["param"] = 1

    return xdata


def check_time_index(index, direct=True):
    if direct:
        name_news = "news_time_index"
        name_air = "day_index"
    else:
        name_news = "NEWS"
        name_air = "air"
    idx = index["news_time"].value[name_news]
    assert len(onp.where(idx > 0)[0].flatten()) == 27

    ref = index["news_time"].time.tolist()
    assert index["news_time"].time.shape == (124, 3)

    assert ref[0] == [NaT, NaT, NaT]
    assert ref[53] == [946872970050000001, 946872980050000001, 946872987050000001]

    ref = index["news_time"].value[name_news].tolist()
    assert ref[0] == [-99999, -99999, -99999]
    assert ref[53] == [2, 3, 4]

    assert index["news_time"].value[name_news].shape == (124, 3)

    ref = index["day"].time.tolist()[20:30]
    assert ref == [
        NaT,
        NaT,
        NaT,
        NaT,
        946684800000000000,
        946684800000000000,
        946684800000000000,
        946684800000000000,
        946684800000000000,
        946684800000000000,
    ]

    ref = index["day"].value[name_air].tolist()[20:30]
    assert (ref == array([-99999, -99999, -99999, -99999, 0, 0, 0, 0, 0, 0])).all()


def check_outputs(outputs, schema):

    # TODO: check what is correct...
    # assert outputs["NEWS"].shape == (25, 3, 2)
    assert outputs["NEWS"].shape == (124, 3)
    assert_tree_all_close(
        outputs["NEWS"][-1],
        DeviceArray([-99999, -99999, -99999], dtype=int32),
    )
    assert_tree_all_close(
        outputs["NEWS"][53],
        DeviceArray([27, 6, 23], dtype=int32),
    )

    assert_tree_all_close(
        outputs["NEWS"][:, 0][60:70],
        DeviceArray(
            [-99999, -99999, 23, -99999, 10, 1, -99999, -99999, -99999, -99999],
            dtype=int32,
        ),
    )

    # decode outputs
    assert outputs["NEWS"].shape == (124, 3)


def test_stream_dataset_dynamic_unroll2():
    dataset = prepare_test_data()
    assert dataset["NEWS"].shape == (14, 2)
    schema = Stream.get_dataset_schema(dataset, check=True)
    ffills = {"day": True}
    freqs = {"day": "d"}
    local_time = "time"
    buffer_maxlen = {"news_time": 3}

    datasets = split_dataset_from_time_dims(dataset)
    assert datasets.keys() == {"day", "news_time", "time"}
    assert datasets["news_time"]["NEWS"].shape == (14, 2)

    stream = Stream(
        local_time,
        freqs,
        ffills,
        buffer_maxlen,
        verbose=True,
        pbar=["time"],
        trace=True,
    )
    streams = {}
    for time_dim, ds in datasets.items():
        streams[time_dim] = stream.start_dataset_stream(ds, time_dim=time_dim)
    trace = stream.merge(streams)
    index = unroll_stream(trace)

    def check_type(data):
        data, treedef = tree_flatten(data)
        for x in data:
            assert isinstance(x, onp.ndarray)

    check_type(index)

    check_time_index(index, False)

    # convert to dict of numpy

    np_data = dict_map(dataset_to_numpy, datasets)

    # encode values
    assert datasets["news_time"]["NEWS"].shape == (14, 2)
    np_data = dict_map(partial(encode_dataset, schema.encoders), np_data)
    assert np_data["news_time"]["NEWS"].shape == (28,)
    # drop times from index structure
    np_index = {dim: var.value for dim, var in index.items()}

    # explicitly convert in np.float32 and int32 berfore jax conversion to avoid jax warnings.
    np_data, np_index = tree_map(onp_half_precision, (np_data, np_index))
    # convert to jax
    np_data, np_index = ep.convert_to_tensors((np_data, np_index), "jax")

    # check data types:
    def check_type(data):
        data, treedef = tree_flatten(data)
        for x in data:
            assert isinstance(x, jnp.DeviceArray)

    check_type(np_data)
    check_type(np_index)

    access_dataset = transform_with_state(partial(tree_access_data, np_data, np_index))

    xs = onp.arange(len(schema.coords["time"]))

    # outputs, state = static_unroll(access_dataset, xs, next(seq))
    # now test dynamic unroll
    outputs, state = unroll(access_dataset, return_final_state=True)(xs)

    check_outputs(outputs["news_time"], schema)


def test_stream_dataset_dynamic_unroll3():
    dataset = prepare_test_data()
    assert dataset["NEWS"].shape == (14, 2)
    schema = Stream.get_dataset_schema(dataset, check=True)
    ffills = {"day": True}
    freqs = {"day": "d"}
    local_time = "time"
    buffer_maxlen = {"news_time": 3}

    time_dataset = get_time_dataset(dataset)

    time_datasets = split_dataset_from_time_dims(time_dataset)
    stream = Stream(
        local_time, freqs, ffills, buffer_maxlen, verbose=["time"], pbar=True
    )
    streams = stream.start_dataset_streams(time_datasets)
    trace = stream.merge(streams)
    time_index = unroll_stream(trace)

    # check time_index
    def check_type(data):
        data, treedef = tree_flatten(data)
        for x in data:
            assert isinstance(x, onp.ndarray)

    check_type(time_index)
    check_time_index(time_index)

    from .stream import get_dataset_index, get_dataset_index_from_stream_index

    # now convert in index for the original dataset.
    time_dataset_index = get_dataset_index_from_stream_index(time_index)
    dataset_index = get_dataset_index(dataset, time_dataset_index)
    assert dataset["param"].shape == ()

    assert dataset_index["param"].shape == (124,)
    # check that dataset_index and dataset have same variables.
    assert dataset.keys() == dataset_index.keys()

    # convert to dict of numpy

    np_data = dataset_to_numpy(dataset)
    np_index = dataset_to_numpy(dataset_index)
    assert np_index["param"].shape == (124,)

    # encode values
    assert dataset["NEWS"].shape == (14, 2)
    np_data_encoded = encode_dataset(schema.encoders, np_data)
    from wax.encode import decode_dataset

    np_data_reencoded = decode_dataset(schema.encoders, np_data_encoded)
    for key in np_data:
        assert (np_data[key] == np_data_reencoded[key]).all()
    np_data = np_data_encoded

    assert np_data["NEWS"].shape == (28,)
    # drop times from index structure

    # explicitly convert in np.float32 and int32 berfore jax conversion to avoid jax warnings.
    np_data, np_index = tree_map(onp_half_precision, (np_data, np_index))
    # convert to jax
    np_data, np_index = ep.convert_to_tensors((np_data, np_index), "jax")

    # check data types:
    def check_type(data):
        data, treedef = tree_flatten(data)
        for x in data:
            assert isinstance(x, jnp.DeviceArray)

    check_type(np_data)
    check_type(np_index)
    assert np_index["param"].shape == (124,)
    access_dataset = transform_with_state(partial(tree_access_data, np_data, np_index))

    xs = onp.arange(len(schema.coords["time"]))

    # outputs, state = static_unroll(access_dataset, xs, next(seq))
    # now test dynamic unroll
    outputs, state = unroll(access_dataset, return_final_state=True)(xs)

    # decode outputs
    assert outputs["NEWS"].shape == (124, 3)
    check_outputs(outputs, schema)



# -*- coding: utf-8 -*-
# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py,md
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.5'
#       jupytext_version: 1.13.0
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# +
# Uncomment to run the notebook in Colab
# # ! pip install -q "wax-ml[complete]@git+https://github.com/eserie/wax-ml.git"
# # ! pip install -q --upgrade jax jaxlib==0.1.70+cuda111 -f https://storage.googleapis.com/jax-releases/jax_releases.html
# -

# check available devices
import jax

print("jax backend {}".format(jax.lib.xla_bridge.get_backend().platform))
jax.devices()

# # 🔭 Reconstructing the light curve of stars with LSTM 🔭
#
# [![Open in Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/github/eserie/wax-ml/blob/main/docs/notebooks/05_reconstructing_the_light_curve_of_stars.ipynb)

# Let's take a walk through the stars...
#
# This notebook is based on the study done in
# [this post by Christophe Pere](https://towardsdatascience.com/how-to-use-deep-learning-for-time-series-forecasting-3f8a399cf205)
# and the notebook available on
# [the authors's github](https://github.com/Christophe-pere/Time_series_RNN).
#
# We will repeat this study on starlight using the LSTM architecture to predict the observed light flux through time.
#
# Our LSTM implementation is based on this [notebook from Haiku's github repository](https://github.com/deepmind/dm-haiku/blob/master/examples/haiku_lstms.ipynb).
#
# We'll see how to use WAX-ML to ease the preparation of time series data stored in dataframes and having Nans
# before calling a "standard" deep-learning workflow.
#
# ## Disclaimer
#
# Despite the fact that this code works with real data, the results presented here should not be considered as scientific knowledge insights, to the knowledge of the authors of WAX-ML, neither the results nor the data source have been reviewed by an astrophysics pair.
#
# The purpose of this notebook is only to demonstrate how WAX-ML can be used when applying a "standard" machine learning workflow, here LSTM, to analyze time series.

# ## Download the data

# %matplotlib inline

import io
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np

# +
import pandas as pd
import requests
from sklearn.preprocessing import MinMaxScaler
from tqdm.auto import tqdm

from wax.accessors import register_wax_accessors
from wax.modules import RollingMean

register_wax_accessors()

# + tags=["parameters"]
# Parameters
STAR = "007609553"
SEQ_LEN = 64
BATCH_SIZE = 8
TRAIN_STEPS = 2 ** 16
TRAIN_SIZE = 2 ** 16
TOTAL_LEN = None
TRAIN_DATE = "2016"
CACHE_DIR = Path("./cached_data/")
# -

# %%time
filename = CACHE_DIR / "kep_lightcurves.parquet"
try:
    raw_dataframe = pd.read_parquet(open(filename, "rb"))
    print(f"data read from {filename}")
except FileNotFoundError:
    # Downloading the csv file from Chrustioge Pere GitHub account
    download = requests.get(
        "https://raw.github.com/Christophe-pere/Time_series_RNN/master/kep_lightcurves.csv"
    ).content
    raw_dataframe = pd.read_csv(io.StringIO(download.decode("utf-8")))
    # set date index
    raw_dataframe.index = pd.Index(
        pd.date_range("2009-03-07", periods=len(raw_dataframe.index), freq="h"),
        name="time",
    )
    # save dataframe locally in CACHE_DIR
    CACHE_DIR.mkdir(exist_ok=True)
    raw_dataframe.to_parquet(filename)
    print(f"data saved in {filename}")


# shortening of data to speed up the execution of the notebook in the CI
if TOTAL_LEN:
    raw_dataframe = raw_dataframe.iloc[:TOTAL_LEN]


# Let's visualize the description of this dataset:

raw_dataframe.describe().T.to_xarray()


stars = raw_dataframe.columns
stars = sorted(list(set([i.split("_")[0] for i in stars])))
print(f"The number of stars available is: {len(stars)}")
print(f"star identifiers: {stars}")

dataframe = raw_dataframe[[i + "_rscl" for i in stars]].rename(
    columns=lambda c: c.replace("_rscl", "")
)
dataframe.columns.names = ["star"]
dataframe.shape

# ## Rolling mean

# We will smooth the data by applying a rolling mean with a window of 100 periods.

# ### Count nan values
#
# But before since the dataset has some nan values, we will extract few statistics
# about the density of nan values in windows of size 100.
#
# It will be the occasion to show a usage of the `wax.modules.Buffer` module with the `format_outputs=False`
# option for the dataframe accessor `.wax.stream`.

import jax.numpy as jnp
import numpy as onp

from wax.modules import Buffer

# Let's apply the `Buffer` module to the data:

buffer, _ = dataframe.wax.stream(format_outputs=False).apply(lambda x: Buffer(100)(x))

assert isinstance(buffer, jnp.ndarray)

# Let's describe the statistic of nans with pandas:

count_nan = jnp.isnan(buffer).sum(axis=1)
pd.DataFrame(onp.array(count_nan)).stack().describe().astype(int)

# ### Computing the rolling mean

# We will choose a `min_periods` of 5 in order to keep at leas 75% of the points.

# %%time
dataframe_mean, _ = dataframe.wax.stream().apply(
    lambda x: RollingMean(100, min_periods=5)(x)
)

dataframe.loc[:, "008241079"].plot()
dataframe_mean.loc[:, "008241079"].plot()

# ### With Dataset API
#
# Let's illustrate how to do the same rolling mean operation but using wax accessors on xarray `Dataset`.

# +
from functools import partial

from jax.tree_util import tree_map

# -

dataset = dataframe.to_xarray()

dataset

# %%time
dataset_mean, _ = dataset.wax.stream().apply(
    partial(tree_map, lambda x: RollingMean(100, min_periods=5)(x)),
    format_dims=["time"],
)

# (Its much longer than with dataframe)
#
# TODO: This is an issue that we should solve.

dataset_mean

dataset["008241079"].plot()
dataset_mean["008241079"].plot()

# ### With dataarray

dataarray = dataframe.to_xarray().to_array("star").transpose("time", "star")

dataarray

# %%time
dataarray_mean, _ = dataarray.wax.stream().apply(
    lambda x: RollingMean(100, min_periods=5)(x)
)

# (Its much longer than with dataframe)

dataarray_mean

dataarray.sel(star="008241079").plot()
dataarray_mean.sel(star="008241079").plot()

# ## Forecasting with Machine Learning
#
# We need two forecast in this data, if you look with attention you'll see micro holes and big holes.


# +
import warnings
from typing import NamedTuple, Tuple, TypeVar

import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np
import optax
import pandas as pd
import plotnine as gg

T = TypeVar("T")
Pair = Tuple[T, T]


class Pair(NamedTuple):
    x: T
    y: T


class TrainSplit(NamedTuple):
    train: T
    validation: T


gg.theme_set(gg.theme_bw())
warnings.filterwarnings("ignore")

# +
import matplotlib.pyplot as plt

plt.rcParams["figure.figsize"] = 18, 8
fig, (ax, lax) = plt.subplots(ncols=2, gridspec_kw={"width_ratios": [4, 1]})
dataframe.plot(ax=ax, title="raw data")
ax.legend(bbox_to_anchor=(0, 0, 1, 1), bbox_transform=lax.transAxes)
lax.axis("off")

# +
import matplotlib.pyplot as plt

plt.rcParams["figure.figsize"] = 18, 8
fig, (ax, lax) = plt.subplots(ncols=2, gridspec_kw={"width_ratios": [4, 1]})
dataframe_mean.plot(ax=ax, title="Smoothed data")
ax.legend(bbox_to_anchor=(0, 0, 1, 1), bbox_transform=lax.transAxes)
lax.axis("off")
# -

# ### Normalize data

dataframe_mean.stack().hist(bins=100)

# +
from wax.encode import Encoder


def min_max_scaler(values: pd.DataFrame, output_format: str = "dataframe") -> Encoder:
    scaler = MinMaxScaler(feature_range=(0, 1))
    scaler.fit(values)
    index = values.index
    columns = values.columns

    def encode(dataframe: pd.DataFrame):
        nonlocal index
        nonlocal columns
        index = dataframe.index
        columns = dataframe.columns
        array_normed = scaler.transform(dataframe)
        if output_format == "dataframe":
            return pd.DataFrame(array_normed, index, columns)
        elif output_format == "jax":
            return jnp.array(array_normed)
        else:
            return array_normed

    def decode(array_scaled):
        value = scaler.inverse_transform(array_scaled)
        if output_format == "dataframe":
            return pd.DataFrame(value, index, columns)
        else:
            return value

    return Encoder(encode, decode)


# -

scaler = min_max_scaler(dataframe_mean)
dataframe_normed = scaler.encode(dataframe_mean)
assert (scaler.decode(dataframe_normed) - dataframe_mean).stack().abs().max() < 1.0e-4

dataframe_normed.stack().hist(bins=100)

# ### Prepare train / validation datasets

from wax.modules import FillNanInf, Lag


# +
def split_feature_target(dataframe, look_back=SEQ_LEN) -> Pair:
    x, _ = dataframe.wax.stream(format_outputs=False).apply(
        lambda x: FillNanInf()(Lag(1)(Buffer(look_back)(x)))
    )
    B, T, F = x.shape
    x = x.transpose(1, 0, 2)

    y, _ = dataframe.wax.stream(format_outputs=False).apply(
        lambda x: FillNanInf()(Buffer(look_back)(x))
    )
    y = y.transpose(1, 0, 2)
    return Pair(x, y)


def split_feature_target(
    dataframe,
    look_back=SEQ_LEN,
    stack=True,
    shuffle=False,
    min_periods_ratio: float = 0.8,
) -> Pair:
    x, _ = dataframe.wax.stream(format_outputs=False).apply(
        lambda x: Lag(1)(Buffer(look_back)(x))
    )
    x = x.transpose(1, 0, 2)

    y, _ = dataframe.wax.stream(format_outputs=False).apply(
        lambda x: Buffer(look_back)(x)
    )
    y = y.transpose(1, 0, 2)

    T, B, F = x.shape

    if stack:
        x = x.reshape(T, B * F, 1)
        y = y.reshape(T, B * F, 1)

    if shuffle:
        rng = jax.random.PRNGKey(42)
        idx = jnp.arange(x.shape[1])
        idx = jax.random.shuffle(rng, idx)
        x = x[:, idx]
        y = y[:, idx]

    if min_periods_ratio:
        count_nan = jnp.isnan(x).sum(axis=0)
        mask = count_nan < min_periods_ratio * T
        idx = jnp.where(mask)
        # print("count_nan = ", count_nan)
        # print("B = ", B)
        x = x[:, idx[0], :]
        y = y[:, idx[0], :]
        T, B, F = x.shape
        # print("B = ", B)

    # round Batch size to a power of to
    B_round = int(2 ** jnp.floor(jnp.log2(B)))
    x = x[:, :B_round, :]
    y = y[:, :B_round, :]

    # fillnan by zeros
    fill_nan_inf = hk.transform(lambda x: FillNanInf()(x))
    params = fill_nan_inf.init(None, jnp.full(x.shape, jnp.nan, x.dtype))
    x = fill_nan_inf.apply(params, None, x)
    y = fill_nan_inf.apply(params, None, y)

    return Pair(x, y)


# -


def split_train_validation(dataframe, stars, train_size, look_back) -> TrainSplit:

    # prepare scaler
    dataframe_train = dataframe[stars].iloc[:train_size]
    scaler = min_max_scaler(dataframe_train)

    # prepare train data
    dataframe_train_normed = scaler.encode(dataframe_train)
    train = split_feature_target(dataframe_train_normed, look_back)

    # prepare validation data
    valid_size = len(dataframe[stars]) - train_size
    valid_size = int(2 ** jnp.floor(jnp.log2(valid_size)))
    valid_end = int(train_size + valid_size)
    dataframe_valid = dataframe[stars].iloc[train_size:valid_end]
    dataframe_valid_normed = scaler.encode(dataframe_valid)
    valid = split_feature_target(dataframe_valid_normed, look_back)

    return TrainSplit(train, valid)


print(f"Look at star: {STAR}")
train, valid = split_train_validation(dataframe_normed, [STAR], TRAIN_SIZE, SEQ_LEN)

train[0].shape, train[1].shape, valid[0].shape, valid[1].shape

TRAIN_SIZE, VALID_SIZE = len(train.x), len(valid.x)


seq = hk.PRNGSequence(42)

# Plot an observation/target pair.
batch_plot = jax.random.choice(next(seq), len(train[0]))
df = pd.DataFrame(
    {"x": train[0][:, batch_plot, 0], "y": train[1][:, batch_plot, 0]}
).reset_index()
df = pd.melt(df, id_vars=["index"], value_vars=["x", "y"])
plot = (
    gg.ggplot(df)
    + gg.aes(x="index", y="value", color="variable")
    + gg.geom_line()
    + gg.scales.scale_y_log10()
)
_ = plot.draw()


# ### Dataset iterator


class Dataset:
    """An iterator over a numpy array, revealing batch_size elements at a time."""

    def __init__(self, xy: Pair, batch_size: int):
        self._x, self._y = xy
        self._batch_size = batch_size
        self._length = self._x.shape[1]
        self._idx = 0
        if self._length % batch_size != 0:
            msg = "dataset size {} must be divisible by batch_size {}."
            raise ValueError(msg.format(self._length, batch_size))

    def __next__(self) -> Pair:
        start = self._idx
        end = start + self._batch_size
        x, y = self._x[:, start:end], self._y[:, start:end]
        if end >= self._length:
            end = end % self._length
            assert end == 0  # Guaranteed by ctor assertion.
        self._idx = end
        return x, y


train_ds = Dataset(train, BATCH_SIZE)
valid_ds = Dataset(valid, BATCH_SIZE)
del train, valid  # Don't leak temporaries.

# + [markdown] colab_type="text" id="LZGw5Jdvjmqh"
# ### Training an LSTM
#
# To train the LSTM, we define a Haiku function which unrolls the LSTM over the input sequence, generating predictions for all output values. The LSTM always starts with its initial state at the start of the sequence.
#
# The Haiku function is then transformed into a pure function through `hk.transform`, and is trained with Adam on an L2 prediction loss.
# -

from wax.compile import jit_init_apply

x, y = next(train_ds)
x.shape, y.shape

from collections import defaultdict


# + colab={} colab_type="code" id="nacnTj5ejIK5"
def unroll_net(seqs: jnp.ndarray):
    """Unrolls an LSTM over seqs, mapping each output to a scalar."""
    # seqs is [T, B, F].
    core = hk.LSTM(32)
    batch_size = seqs.shape[1]
    outs, state = hk.dynamic_unroll(core, seqs, core.initial_state(batch_size))
    # We could include this Linear as part of the recurrent core!
    # However, it's more efficient on modern accelerators to run the linear once
    # over the entire sequence than once per sequence element.
    return hk.BatchApply(hk.Linear(1))(outs), state


# + colab={} colab_type="code" id="nacnTj5ejIK5"
model = jit_init_apply(hk.transform(unroll_net))


# + colab={} colab_type="code" id="nacnTj5ejIK5"
def train_model(
    train_ds: Dataset, valid_ds: Dataset, max_iterations: int = -1
) -> hk.Params:
    """Initializes and trains a model on train_ds, returning the final params."""
    rng = jax.random.PRNGKey(428)
    opt = optax.adam(1e-3)

    @jax.jit
    def loss(params, x, y):
        pred, _ = model.apply(params, None, x)
        return jnp.mean(jnp.square(pred - y))

    @jax.jit
    def update(step, params, opt_state, x, y):
        l, grads = jax.value_and_grad(loss)(params, x, y)
        grads, opt_state = opt.update(grads, opt_state)
        params = optax.apply_updates(params, grads)
        return l, params, opt_state

    # Initialize state.
    sample_x, _ = next(train_ds)
    params = model.init(rng, sample_x)
    opt_state = opt.init(params)

    step = 0
    records = defaultdict(list)

    def _format_results(records):
        records = {key: jnp.stack(l) for key, l in records.items()}
        return records

    with tqdm() as pbar:
        while True:
            if step % 100 == 0:
                x, y = next(valid_ds)
                valid_loss = loss(params, x, y)
                # print("Step {}: valid loss {}".format(step, valid_loss))
                records["step"].append(step)
                records["valid_loss"].append(valid_loss)

            try:
                x, y = next(train_ds)
            except StopIteration:
                return params, _format_results(records)
            train_loss, params, opt_state = update(step, params, opt_state, x, y)
            if step % 100 == 0:
                # print("Step {}: train loss {}".format(step, train_loss))
                records["train_loss"].append(train_loss)

            step += 1
            pbar.update()
            if max_iterations > 0 and step >= max_iterations:
                return params, _format_results(records)


# + colab={} colab_type="code" id="AssgDctokbl5"
# %%time
trained_params, records = train_model(train_ds, valid_ds, TRAIN_STEPS)
# -

# Plot losses
losses = pd.DataFrame(records)
df = pd.melt(losses, id_vars=["step"], value_vars=["train_loss", "valid_loss"])
plot = (
    gg.ggplot(df)
    + gg.aes(x="step", y="value", color="variable")
    + gg.geom_line()
    + gg.scales.scale_y_log10()
)
_ = plot.draw()


# + [markdown] colab_type="text" id="yr7jrOL3ki-b"
# ### Sampling
#
# The point of training models is so that they can make predictions! How can we generate predictions with the trained model?
#
# If we're allowed to feed in the ground truth, we can just run the original model's `apply` function.

# + colab={} colab_type="code" id="f2qETEqXLT1N"
def plot_samples(truth: np.ndarray, prediction: np.ndarray) -> gg.ggplot:
    assert truth.shape == prediction.shape
    df = pd.DataFrame(
        {"truth": truth.squeeze(), "predicted": prediction.squeeze()}
    ).reset_index()
    df = pd.melt(df, id_vars=["index"], value_vars=["truth", "predicted"])
    plot = (
        gg.ggplot(df) + gg.aes(x="index", y="value", color="variable") + gg.geom_line()
    )
    return plot


# + colab={} colab_type="code" id="KOuK1egilGD0"
# Grab a sample from the validation set.
sample_x, _ = next(valid_ds)
sample_x = sample_x[:, :1]  # Shrink to batch-size 1.

# Generate a prediction, feeding in ground truth at each point as input.
predicted, _ = model.apply(trained_params, None, sample_x)

plot = plot_samples(sample_x[1:], predicted[:-1])
plot.draw()
del sample_x, predicted


# -

# ### Run autoregressively

# + [markdown] colab_type="text" id="tDyGshz_lwrM"
# If we can't feed in the ground truth (because we don't have it), we can also run the model autoregressively.

# + colab={} colab_type="code" id="Cg8oQ75Ulvld"
def autoregressive_predict(
    trained_params: hk.Params,
    context: jnp.ndarray,
    seq_len: int,
):
    """Given a context, autoregressively generate the rest of a sine wave."""
    ar_outs = []
    context = jax.device_put(context)
    times = range(seq_len - context.shape[0])
    for _ in times:
        full_context = jnp.concatenate([context] + ar_outs)
        outs, _ = jax.jit(model.apply)(trained_params, None, full_context)
        # Append the newest prediction to ar_outs.
        ar_outs.append(outs[-1:])
    # Return the final full prediction.
    return outs


# + colab={} colab_type="code" id="Cg8oQ75Ulvld"
sample_x, _ = next(valid_ds)
context_length = SEQ_LEN // 8
# Cut the batch-size 1 context from the start of the sequence.
context = sample_x[:context_length, :1]

# + colab={} colab_type="code" id="Cg8oQ75Ulvld"
# %%time
# We can reuse params we got from training for inference - as long as the
# declaration order is the same.
predicted = autoregressive_predict(trained_params, context, SEQ_LEN)

plot = plot_samples(sample_x[1:, :1], predicted)
plot += gg.geom_vline(xintercept=len(context), linetype="dashed")
plot.draw()
del predicted


# + [markdown] colab_type="text" id="qGkr2gf2oALo"
# #### Sharing parameters with a different function.
#
# Unfortunately, this is a bit slow - we're doing O(N^2) computation for a sequence of length N.
#
# It'd be better if we could do the autoregressive sampling all at once - but we need to write a new Haiku function for that.
#
# We're in luck - if the Haiku module names match, the same parameters can be used for multiple Haiku functions.
#
# This can be achieved through a combination of two techniques:
#
# 1. If we manually give a unique name to a module, we can ensure that the parameters are directed to the right places.
# 2. If modules are instantiated in the same order, they'll have the same names in different functions.
#
# Here, we rely on method #2 to create a fast autoregressive prediction.

# + colab={} colab_type="code" id="WdKcHr6_n_ba"
def fast_autoregressive_predict_fn(context, seq_len):
    """Given a context, autoregressively generate the rest of a sine wave."""
    core = hk.LSTM(32)
    dense = hk.Linear(1)
    state = core.initial_state(context.shape[1])
    # Unroll over the context using `hk.dynamic_unroll`.
    # As before, we `hk.BatchApply` the Linear for efficiency.
    context_outs, state = hk.dynamic_unroll(core, context, state)
    context_outs = hk.BatchApply(dense)(context_outs)

    # Now, unroll one step at a time using the running recurrent state.
    ar_outs = []
    x = context_outs[-1]
    times = range(seq_len - context.shape[0])
    for _ in times:
        x, state = core(x, state)
        x = dense(x)
        ar_outs.append(x)
    return jnp.concatenate([context_outs, jnp.stack(ar_outs)])


fast_ar_predict = hk.transform(fast_autoregressive_predict_fn)
fast_ar_predict = jax.jit(fast_ar_predict.apply, static_argnums=3)

# + colab={} colab_type="code" id="WdKcHr6_n_ba"
# %%time
# Reuse the same context from the previous cell.
predicted = fast_ar_predict(trained_params, None, context, SEQ_LEN)

# The plots should be equivalent!
plot = plot_samples(sample_x[1:, :1], predicted[:-1])
plot += gg.geom_vline(xintercept=len(context), linetype="dashed")
_ = plot.draw()


# + colab={} colab_type="code" id="9S0tkPXGrU3a"
# %timeit autoregressive_predict(trained_params, context, SEQ_LEN)
# %timeit fast_ar_predict(trained_params, None, context, SEQ_LEN)
# -
# ## Train all stars


# ### Training


def split_train_validation_date(dataframe, stars, date, look_back) -> TrainSplit:
    train_size = len(dataframe.loc[:date])
    return split_train_validation(dataframe, stars, train_size, look_back)


# %%time
train, valid = split_train_validation_date(dataframe_normed, stars, TRAIN_DATE, SEQ_LEN)
TRAIN_SIZE = train[0].shape[1]
print(f"TRAIN_SIZE = {TRAIN_SIZE}")

train[0].shape, train[1].shape, valid[0].shape, valid[1].shape

train_ds = Dataset(train, BATCH_SIZE)
valid_ds = Dataset(valid, BATCH_SIZE)
del train, valid  # Don't leak temporaries.

# %%time
trained_params, records = train_model(train_ds, valid_ds, TRAIN_STEPS)

# Plot losses
losses = pd.DataFrame(records)
df = pd.melt(losses, id_vars=["step"], value_vars=["train_loss", "valid_loss"])
plot = (
    gg.ggplot(df)
    + gg.aes(x="step", y="value", color="variable")
    + gg.geom_line()
    + gg.scales.scale_y_log10()
)
_ = plot.draw()

# ### Sampling

# +
# Grab a sample from the validation set.
sample_x, _ = next(valid_ds)
sample_x = sample_x[:, :1]  # Shrink to batch-size 1.

# Generate a prediction, feeding in ground truth at each point as input.
predicted, _ = model.apply(trained_params, None, sample_x)

plot = plot_samples(sample_x[1:], predicted[:-1])
plot.draw()
del sample_x, predicted
# -

# ### Run autoregressively

# +
# %%time
sample_x, _ = next(valid_ds)
context_length = SEQ_LEN // 8
# Cut the batch-size 1 context from the start of the sequence.
context = sample_x[:context_length, :1]

# Reuse the same context from the previous cell.
predicted = fast_ar_predict(trained_params, None, context, SEQ_LEN)

# The plots should be equivalent!
plot = plot_samples(sample_x[1:, :1], predicted[:-1])
plot += gg.geom_vline(xintercept=len(context), linetype="dashed")
_ = plot.draw()



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Define accessors for xarray and pandas data containers."""
from dataclasses import dataclass, field
from typing import Any, Callable, Tuple, Union

import jax.numpy as jnp
import numpy as onp
import pandas as pd
import xarray as xr
from jax.tree_util import tree_map

from wax.format import format_dataarray, format_dataframe, format_dataset, format_series
from wax.stream import Stream

# DTypeLike = TypeVar("DTypeLike")
DTypeLike = str


@dataclass(frozen=True)
class StreamDataset(Stream):
    """
    Args:
        accessor : Dataset accessor
    """

    accessor: xr.Dataset = field(default_factory=xr.Dataset)

    def apply(
        self,
        module: Callable,
        skip_first: bool = False,
        format_dims: Union[Tuple, onp.ndarray] = (),
        params: Any = None,
        state: Any = None,
        rng: jnp.ndarray = None,
    ) -> Any:
        """Apply a module to a dataset.

        Args:
            module : callable being able to be transformed with Haiku transform_with_state.
            skip_first : if true, first value of the sequence is not used in apply.
            format_dims : nested data structure with specification of dims for dataarray formatting.
            params: parameters for the module.
            state : state for the module.
            rng: random number generator key.

        Return:
            Unroll results of the module formated as a nested data structure with dataarray leaves.
        """
        dataset = self.accessor._obj
        schema = self.get_dataset_schema(dataset)
        transform_dataset, xs = self.prepare(
            dataset, module, schema.encoders, skip_first=skip_first
        )

        init_params, init_state = transform_dataset.init(rng, xs)
        params = params if params is not None else init_params
        state = state if state is not None else init_state
        outputs, state = transform_dataset.apply(params, state, rng, xs)

        if self.format_outputs:
            outputs = tree_map(lambda x: onp.array(x), outputs)
            try:
                outputs = format_dataset(schema.coords, outputs, format_dims)
            except ValueError:
                # try with reduced dims
                format_dims = (format_dims[0],)
                outputs = format_dataset(schema.coords, outputs, format_dims)

        if self.return_state:
            return outputs, state
        else:
            return outputs


@dataclass(frozen=True)
class StreamDataArray(Stream):
    """
    Args:
        accessor : DataArray accessor
    """

    accessor: xr.Dataset = field(default_factory=xr.Dataset)

    def apply(
        self,
        module: Callable,
        skip_first: bool = False,
        format_dims: Any = None,
        params: Any = None,
        state: Any = None,
        rng: jnp.ndarray = None,
    ) -> Any:
        """Apply a module to a dataset.

        Args:
            module : callable being able to be transformed with Haiku transform_with_state.
            skip_first : if true, first value of the sequence is not used in apply.
            format_dims : nested data structure with specification of dims for dataarray formatting.
            params: parameters for the module.
            state : state for the module.
            rng: random number generator key.

        Return:
            Unroll results of the module formated as a nested data structure with dataarray leaves.
        """
        dataarray = self.accessor._obj
        dataset = xr.Dataset({"dataarray": dataarray})
        schema = self.get_dataset_schema(dataset)

        def module_dataset(dataset):
            array = dataset["dataarray"]
            return module(array)

        transform_dataset, xs = self.prepare(
            dataset, module_dataset, schema.encoders, skip_first=skip_first
        )

        init_params, init_state = transform_dataset.init(rng, xs)
        params = params if params is not None else init_params
        state = state if state is not None else init_state
        outputs, state = transform_dataset.apply(params, state, rng, xs)

        if self.format_outputs:
            outputs = tree_map(lambda x: onp.array(x), outputs)
            if format_dims is None:
                format_dims = dataarray.dims

            try:
                outputs = format_dataarray(schema.coords, outputs, format_dims)
            except ValueError:
                # try with reduced dims
                format_dims = (format_dims[0],)
                outputs = format_dataarray(schema.coords, outputs, format_dims)

        if self.return_state:
            return outputs, state
        else:
            return outputs


@dataclass(frozen=True)
class StreamDataFrame(Stream):
    """
    Args:
        accessor : DataArray accessor
    """

    accessor: xr.Dataset = field(default_factory=xr.Dataset)

    def apply(
        self,
        module: Callable,
        skip_first: bool = False,
        format_dims: Any = None,
        params: Any = None,
        state: Any = None,
        rng: jnp.ndarray = None,
    ) -> Any:
        """Apply a module to a dataset.

        Args:
            module : callable being able to be transformed with Haiku transform_with_state.
            skip_first : if true, first value of the sequence is not used in apply.
            format_dims : nested data structure with specification of dims for dataarray formatting.
            params: parameters for the module.
            state : state for the module.
            rng: random number generator key.

        Return:
            Unroll results of the module formated as a nested data structure with dataarray leaves.
        """
        dataframe = self.accessor._obj

        def _has_multi_index(dataframe):
            if dataframe.index.nlevels > 1:
                return True
            if dataframe.columns.nlevels > 1:
                return True
            return False

        if not _has_multi_index(dataframe):
            # avoid to stack the dataframe : its faster!
            dataarray = xr.DataArray(dataframe)
        else:
            dataarray = dataframe.stack(
                list(range(dataframe.columns.nlevels))
            ).to_xarray()
        dataset = dataarray.to_dataset(name="dataarray")
        schema = self.get_dataset_schema(dataset)

        def module_dataset(dataset):
            array = dataset["dataarray"]
            return module(array)

        transform_dataset, xs = self.prepare(
            dataset, module_dataset, schema.encoders, skip_first=skip_first
        )
        init_params, init_state = transform_dataset.init(rng, xs)
        params = params if params is not None else init_params
        state = state if state is not None else init_state
        outputs, state = transform_dataset.apply(params, state, rng, xs)

        if self.format_outputs:
            outputs = tree_map(lambda x: onp.array(x), outputs)

            if format_dims is None:
                format_dims = dataarray.dims

            try:
                outputs = format_dataframe(
                    schema.coords,
                    outputs,
                    format_dims,
                    index_nlevels=dataframe.index.nlevels,
                )
            except ValueError:
                # try with reduced dims
                format_dims = (format_dims[0],)
                outputs = format_dataframe(
                    schema.coords,
                    outputs,
                    format_dims,
                    index_nlevels=dataframe.index.nlevels,
                )

        if self.return_state:
            return outputs, state
        else:
            return outputs


@dataclass(frozen=True)
class StreamSeries(Stream):
    """
    Args:
        accessor : DataArray accessor
    """

    accessor: xr.Dataset = field(default_factory=xr.Dataset)

    def apply(
        self,
        module: Callable,
        skip_first: bool = False,
        format_dims: Any = None,
        params: Any = None,
        state: Any = None,
        rng: jnp.ndarray = None,
    ) -> Any:
        """Apply a module to a dataset.

        Args:
            module : callable being able to be transformed with Haiku transform_with_state.
            skip_first : if true, first value of the sequence is not used in apply.
            format_dims : nested data structure with specification of dims for dataarray formatting.
            params: parameters for the module.
            state : state for the module.
            rng: random number generator key.

        Return:
            Unroll results of the module formated as a nested data structure with dataarray leaves.
        """
        series = self.accessor._obj
        dataarray = series.to_xarray()
        dataset = xr.Dataset({"dataarray": dataarray})
        schema = self.get_dataset_schema(dataset)

        def module_dataset(dataset):
            array = dataset["dataarray"]
            return module(array)

        transform_dataset, xs = self.prepare(
            dataset, module_dataset, schema.encoders, skip_first=skip_first
        )
        init_params, init_state = transform_dataset.init(rng, xs)
        params = params if params is not None else init_params
        state = state if state is not None else init_state
        outputs, state = transform_dataset.apply(params, state, rng, xs)

        if self.format_outputs:
            outputs = tree_map(lambda x: onp.array(x), outputs)
            if format_dims is None:
                format_dims = dataarray.dims

            try:
                outputs = format_series(schema.coords, outputs, format_dims)
            except ValueError:
                # try with reduced dims
                format_dims = (format_dims[0],)
                outputs = format_series(schema.coords, outputs, format_dims)
        if self.return_state:
            return outputs, state
        else:
            return outputs


class WaxAccessor:
    def ewm(self, *args, **kwargs):
        return ExponentialMovingWindow(accessor=self, *args, **kwargs)


@dataclass(frozen=True)
class ExponentialMovingWindow:
    accessor: WaxAccessor
    alpha: float
    adjust: bool = True
    return_state: bool = False
    format_outputs: bool = True

    def mean(self):
        from wax.modules import EWMA

        def _apply_ema(
            accessor, alpha, adjust, params=None, state=None, *args, **kwargs
        ):
            return accessor.stream(*args, **kwargs).apply(
                lambda x: EWMA(alpha, adjust)(x),
                params=params,
                state=state,
                rng=None,
            )

        return _apply_ema(**self.__dict__)


class WaxDatasetAccessor(WaxAccessor):
    def __init__(self, xarray_obj):
        self._obj = xarray_obj

    def stream(self, *args, **kwargs):
        return StreamDataset(accessor=self, *args, **kwargs)


class WaxDataArrayAccessor(WaxAccessor):
    def __init__(self, xarray_obj):
        self._obj = xarray_obj

    def stream(self, *args, **kwargs):
        return StreamDataArray(accessor=self, *args, **kwargs)


class WaxDataFrameAccessor(WaxAccessor):
    def __init__(self, pandas_obj):
        self._obj = pandas_obj

    def stream(self, *args, **kwargs):
        return StreamDataFrame(accessor=self, *args, **kwargs)


class WaxSeriesAccessor(WaxAccessor):
    def __init__(self, pandas_obj):
        self._obj = pandas_obj

    def stream(self, *args, **kwargs):
        return StreamSeries(accessor=self, *args, **kwargs)


def register_wax_accessors():
    if not hasattr(xr.Dataset, "wax"):
        xr.register_dataset_accessor("wax")(WaxDatasetAccessor)
    if not hasattr(xr.DataArray, "wax"):
        xr.register_dataarray_accessor("wax")(WaxDataArrayAccessor)
    if not hasattr(pd.DataFrame, "wax"):
        pd.api.extensions.register_dataframe_accessor("wax")(WaxDataFrameAccessor)
    if not hasattr(pd.Series, "wax"):
        pd.api.extensions.register_series_accessor("wax")(WaxSeriesAccessor)



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from .jax import JAXTensor  # noqa: F401
from .numpy import NumPyTensor  # noqa: F401
from .pytorch import PyTorchTensor  # noqa: F401
from .tensor import Tensor  # noqa: F401
from .tensor import TensorOrScalar  # noqa: F401
from .tensor import TensorType  # noqa: F401
from .tensor import istensor  # noqa: F401
from .tensorflow import TensorFlowTensor  # noqa: F401



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from .tensor import TensorType


def kl_div_with_logits(
    logits_p: TensorType, logits_q: TensorType, axis: int = -1, keepdims: bool = False
) -> TensorType:
    log_p = logits_p.log_softmax(axis=axis)
    log_q = logits_q.log_softmax(axis=axis)
    p = logits_p.softmax(axis=-1)
    return (p * (log_p - log_q)).sum(axis=axis, keepdims=keepdims)



# -*- coding: utf-8 -*-
# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py,md
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.5'
#       jupytext_version: 1.13.0
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# +
# Uncomment to run the notebook in Colab
# # ! pip install -q "wax-ml[complete]@git+https://github.com/eserie/wax-ml.git"
# # ! pip install -q --upgrade jax jaxlib==0.1.70+cuda111 -f https://storage.googleapis.com/jax-releases/jax_releases.html
# -

# check available devices
import jax

print("jax backend {}".format(jax.lib.xla_bridge.get_backend().platform))
jax.devices()

# # 🌡 Binning temperatures 🌡
#
# [![Open in Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/github/eserie/wax-ml/blob/main/docs/notebooks/03_ohlc_temperature.ipynb)

# Let's again considering the air temperatures dataset.
# It is sampled at an hourly resolution.
# We will make "trailing" air temperature bins during each day and "reset" the bin
# aggregation process at each day change.

# + tags=[]
import numpy as onp
import xarray as xr

# + tags=[]
from wax.accessors import register_wax_accessors
from wax.modules import OHLC, HasChanged

register_wax_accessors()

# + tags=[]
dataset = xr.tutorial.open_dataset("air_temperature")
dataset["date"] = dataset.time.dt.date.astype(onp.datetime64)
# -

dataset


# + tags=[]
def bin_temperature(da):
    day_change = HasChanged()(da["date"])
    return OHLC()(da["air"], reset_on=day_change)


output, state = dataset.wax.stream().apply(
    bin_temperature, format_dims=onp.array(dataset.air.dims)
)
output = xr.Dataset(output._asdict())

# + tags=[]
df = output.isel(lat=0, lon=0).drop(["lat", "lon"]).to_pandas().loc["2013-01"]
_ = df.plot(figsize=(12, 8), title="Trailing Open-High-Low-Close temperatures")

# + [markdown] tags=[]
# ## The `UpdateOnEvent` module
#
# The `OHLC` module uses the primitive `wax.modules.UpdateOnEvent`.
#
# Its implementation required to complete Haiku with a central function
# `set_params_or_state_dict` which we have actually integrated in this WAX-ML module.
#
# We have opened an [issue on the Haiku github](https://github.com/deepmind/dm-haiku/issues/126)
# to integrate it in Haiku.



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Gym feedback between an agent and a Gym environment."""
from collections import namedtuple
from dataclasses import dataclass, field
from typing import Any, NamedTuple

import haiku as hk


@dataclass
class GymOutput:
    """Dynamically define GymOutput namedtuple structure.

    Args:
        return_obs: if true, add 'obs' field to output structure.
        return_action: if true, action 'obs' field to output structure.
    """

    return_obs: bool = False
    return_action: bool = False
    gym_output_struct: Any = field(init=False)

    def __post_init__(self):
        outputs = ["reward"]
        if self.return_obs:
            outputs += ["obs"]
        if self.return_action:
            outputs += ["action"]
        self.gym_output_struct = namedtuple("GymOutput", outputs)

    def format(self, reward, obs=None, action=None):
        outputs = [reward]
        if self.return_obs:
            outputs += [obs]
        if self.return_action:
            outputs += [action]
        return self.gym_output_struct(*outputs)


class GymInfo(NamedTuple):
    env: Any
    agent: Any


class GymFeedback(hk.Module):
    """Gym feedback between an agent and a Gym environment."""

    GymState = namedtuple("GymState", "action")

    def __init__(
        self,
        agent,
        env,
        return_obs=False,
        return_action=False,
        init_action=None,
        name=None,
    ):
        """Initialize module.

        Args:
            agent : Gym environment used to unroll the data and feed the agent.
            env : Gym environment used to unroll the data and feed the agent.
            return_obs : if true return environment observation
            return_action : if true return agent action
            name : name of the module

        """
        super().__init__(name=name)
        self.agent = agent
        self.env = env
        self.return_obs = return_obs
        self.return_action = return_action
        self.init_action = init_action

        self.GymOutput = GymOutput(self.return_obs, self.return_action)

    def __call__(self, raw_obs):
        """Compute Gym feedback loop.

        Args:
            raw_obs: raw observations.
        Returns:
            gym_output : reward.
                Use return_obs=True to also return env observations.
                Use return_action=True to aslo return agent actions.
        """

        def init_action(shape, dtype):
            if self.init_action is not None:
                return self.GymState(self.init_action)
            else:
                # call
                return self.GymState(self.agent(raw_obs)[0])

        action = hk.get_state("action", shape=[], init=init_action)

        rw, obs, info_env = self.env(action.action, raw_obs)
        action, info_agent = self.agent(obs)

        info = GymInfo(info_env, info_agent)
        outputs = self.GymOutput.format(rw, obs, action)
        state = self.GymState(action)
        hk.set_state("action", state)
        return outputs, info



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from functools import partial

# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, NamedTuple

import haiku as hk
import jax
import jax.numpy as jnp
import numpy as onp
import pandas as pd
import pytest
from optax._src.base import OptState

from wax.modules import ARMA, SNARIMAX, GymFeedback, OnlineOptimizer, UpdateParams, VMap
from wax.modules.lag import tree_lag
from wax.modules.vmap import add_batch
from wax.optim.newton import newton
from wax.unroll import unroll_transform_with_state


def generate_arma():
    alpha = jnp.array([0.6, -0.5, 0.4, -0.4, 0.3])
    beta = jnp.array([0.3, -0.2])

    rng = jax.random.PRNGKey(42)
    eps = jax.random.normal(rng, (100,))
    sim = unroll_transform_with_state(lambda eps: ARMA(alpha, beta)(eps))
    params, state = sim.init(rng, eps)
    y, state = sim.apply(params, state, rng, eps)
    return y


def test_snarimax_not_implemented():
    sim = hk.transform_with_state(lambda y: SNARIMAX(1, 1, 0)(y))
    rng = jax.random.PRNGKey(42)

    y = jax.random.normal(rng, (1,))
    with pytest.raises(NotImplementedError):
        sim.init(rng, y)


def test_snarimax():
    y = generate_arma()

    def predict(y, X=None):
        return SNARIMAX(10, 0, 0)(y, X)

    sim = unroll_transform_with_state(predict)
    rng = jax.random.PRNGKey(42)
    params, state = sim.init(rng, y)
    (y_pred, _), state = sim.apply(params, state, rng, y)

    assert y_pred.shape == y.shape


def build_env():
    def env(action, obs):
        y_pred, eps = action, obs
        ar_coefs = jnp.array([0.6, -0.5, 0.4, -0.4, 0.3])
        ma_coefs = jnp.array([0.3, -0.2])

        y = ARMA(ar_coefs, ma_coefs)(eps)
        # prediction used on a fresh y observation.
        rw = -((y - y_pred) ** 2)

        env_info = {"y": y, "y_pred": y_pred}
        obs = y
        return rw, obs, env_info

    return env


def build_agent(time_series_model=None, opt=None):
    if time_series_model is None:

        def time_series_model(y, X):
            return SNARIMAX(10)(y, X)

    if opt is None:
        opt = newton(0.3, 0.3)

    class AgentInfo(NamedTuple):
        optim: Any
        forecast: Any

    class ModelWithLossInfo(NamedTuple):
        pred: Any
        loss: Any

    def agent(obs):
        if isinstance(obs, tuple):
            y, X = obs
        else:
            y = obs
            X = None

        def evaluate(y_pred, y):
            return jnp.linalg.norm(y_pred - y) ** 2, {}

        def model_with_loss(y, X=None):
            # predict with lagged data
            y_pred, pred_info = time_series_model(*tree_lag(1)(y, X))

            # evaluate loss with actual data
            loss, loss_info = evaluate(y_pred, y)

            return loss, ModelWithLossInfo(pred_info, loss_info)

        def project_params(params: Any, opt_state: OptState = None):
            del opt_state
            return jax.tree_map(lambda w: jnp.clip(w, -1, 1), params)

        def split_params(params):
            def filter_params(m, n, p):
                # print(m, n, p)
                return m.endswith("snarimax/~/linear") and n == "w"

            return hk.data_structures.partition(filter_params, params)

        def learn_and_forecast(y, X=None):
            optim_res = OnlineOptimizer(
                model_with_loss,
                opt,
                project_params=project_params,
                split_params=split_params,
            )(*tree_lag(1)(y, X))

            predict_params = optim_res.updated_params

            y_pred, forecast_info = UpdateParams(time_series_model)(
                predict_params, y, X
            )
            return y_pred, AgentInfo(optim_res, forecast_info)

        return learn_and_forecast(y, X)

    return agent


def run_scan_hyper_params(env):
    n_time_step = 100
    n_batches = 3
    n_step_size = 2
    n_eps = 2

    STEP_SIZE = pd.Index(onp.logspace(-2, 3, n_step_size), name="step_size")
    EPS = pd.Index(onp.logspace(-4, 3, n_eps), name="eps")

    HPARAMS_idx = pd.MultiIndex.from_product([STEP_SIZE, EPS])
    HPARAMS = jnp.stack(list(map(onp.array, HPARAMS_idx)))

    @partial(add_batch, take_mean=False)
    def gym_loop_scan_hparams(eps):
        def scan_params(hparams):
            step_size, newton_eps = hparams
            agent = build_agent(opt=newton(step_size, eps=newton_eps))
            return GymFeedback(agent, env)(eps)

        return VMap(scan_params)(HPARAMS)

    sim = unroll_transform_with_state(gym_loop_scan_hparams)
    rng = jax.random.PRNGKey(42)
    eps = jax.random.normal(rng, (n_time_step, n_batches)) * 0.3

    params, state = sim.init(rng, eps)
    (gym, info), state = sim.apply(params, state, rng, eps)

    assert gym.reward.shape == (n_time_step, n_batches, n_step_size * n_eps)


def test_gym_loop():
    env = build_env()
    agent = build_agent()

    rng = jax.random.PRNGKey(42)
    eps = jax.random.normal(rng, (100,)) * 0.3

    sim = unroll_transform_with_state(lambda eps: GymFeedback(agent, env)(eps))
    params, state = sim.init(rng, eps)
    (gym, info), state = sim.apply(params, state, rng, eps)

    # pd.Series(-gym.reward).expanding().mean().plot(label="AR(10) model")
    assert -gym.reward.sum() > 0


def test_scan_hyper_params():
    env = build_env()
    run_scan_hyper_params(env)



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from haiku._src.data_structures import FlatMapping
from jax import numpy as jnp
from jax._src.numpy.lax_numpy import array as DeviceArray
from jax._src.numpy.lax_numpy import uint32

from wax.modules.counter import Counter
from wax.testing import assert_tree_all_close
from wax.unroll import unroll


def test_counter():
    res, state = unroll(lambda _: Counter()(), return_final_state=True)(
        jnp.arange(10, 20),
    )
    assert_tree_all_close(res, jnp.arange(1, 11, dtype="uint32"))
    assert_tree_all_close(
        state,
        FlatMapping({"counter": FlatMapping({"count": DeviceArray(10, dtype=uint32)})}),
    )



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Apply a module when an event occur otherwise return last computed output.

Part of this module may be integrated in Haiku.
See: https://github.com/deepmind/dm-haiku/issues/126
"""
from typing import Mapping, Set

import haiku as hk
import jax.numpy as jnp
from haiku._src import base
from jax.tree_util import tree_map


def set_params_or_state_dict(
    module_name: str,
    submodules: Set[str],
    which: str,
    state_dict: Mapping[str, jnp.ndarray],
):
    """Returns module parameters or state for the given module or submodules."""
    assert which in ("params", "state")
    frame = base.current_frame()
    for their_module_name, bundle in getattr(frame, which).items():
        if (
            their_module_name == module_name
            or their_module_name.startswith(module_name + "/")
            or their_module_name in submodules
        ):
            for name, value in bundle.items():
                fq_name = their_module_name + "/" + name
                if which == "state":
                    value_dict = value._asdict()
                    value_dict["current"] = state_dict[fq_name]
                    bundle[name] = type(value)(**value_dict)
                else:
                    bundle[name] = state_dict[fq_name]


def set_state_from_dict(self, next_state_dict):
    """Set state keyed by name for this module and submodules."""
    if not base.frame_stack:
        raise ValueError(
            "`module.set_state_from_dict()` must be used as part of an `hk.transform`."
        )
    set_params_or_state_dict(
        self.module_name,
        self._submodules,
        "state",
        next_state_dict,
    )


hk.Module.set_state_from_dict = set_state_from_dict


class UpdateOnEvent(hk.Module):
    """Apply a module when an event occur otherwise return last computed output.

    If the module has state management, then it will be ask to delegate the state
    management to this module.
    """

    def __init__(self, module, initial_output_value=jnp.nan, name=None):
        super().__init__(name=name)
        self.module = module
        self.initial_output_value = initial_output_value

    def __call__(self, on_event, *args, **kwargs):
        # state_dict = self.state_dict()

        prev_state_dict = self.module.state_dict()

        output = self.module(*args, **kwargs)

        prev_output = hk.get_state(
            "prev_output",
            shape=[],
            init=lambda *_: tree_map(
                lambda x: jnp.full(x.shape, self.initial_output_value, dtype=x.dtype),
                output,
            ),
        )

        next_state_dict = self.module.state_dict() if prev_state_dict else {}

        def true_fun(operand):
            output, next_state, _, _ = operand
            return output, next_state

        def false_fun(operand):
            _, _, prev_output, prev_state = operand
            return prev_output, prev_state

        operand = (output, next_state_dict, prev_output, prev_state_dict)
        output, next_state_dict = hk.cond(
            pred=on_event,
            true_operand=operand,
            true_fun=true_fun,
            false_operand=operand,
            false_fun=false_fun,
        )

        if next_state_dict:
            self.module.set_state_from_dict(next_state_dict)

        hk.set_state("prev_output", output)
        return output



# -*- coding: utf-8 -*-
# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py,md
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.5'
#       jupytext_version: 1.13.0
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# +
# Uncomment to run the notebook in Colab
# # ! pip install -q "wax-ml[complete]@git+https://github.com/eserie/wax-ml.git"
# # ! pip install -q --upgrade jax jaxlib==0.1.70+cuda111 -f https://storage.googleapis.com/jax-releases/jax_releases.html
# -

# check available devices
import jax

print("jax backend {}".format(jax.lib.xla_bridge.get_backend().platform))
jax.devices()

# # ⏱ Synchronize data streams ⏱
#
# [![Open in Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/github/eserie/wax-ml/blob/main/docs/notebooks/02_Synchronize_data_streams.ipynb)

# Physicists, and not the least 😅, have brought a solution to the synchronization
# problem.  See [Poincaré-Einstein synchronization Wikipedia
# page](https://en.wikipedia.org/wiki/Einstein_synchronisation) for more details.
#
# In WAX-ML we strive to follow their recommendations and implement a synchronization
# mechanism between different data streams. Using the terminology of Henri Poincaré (see
# link above), we introduce the notion of "local time" to unravel the stream in which
# the user wants to apply transformations. We call the other streams "secondary streams".
# They can work at different frequencies, lower or higher.  The data from these secondary
# streams will be represented in the "local time" either with the use of a
# forward filling mechanism for lower frequencies or a buffering mechanism
# for higher frequencies.
#
# We implement a "data tracing" mechanism to optimize access to out-of-sync streams.
# This mechanism works on in-memory data.  We perform the first pass on the data,
# without actually accessing it, and determine the indices necessary to
# later access the data. Doing so we are vigilant to not let any "future"
# information pass through and thus guaranty a data processing that respects causality.
#
# The buffering mechanism used in the case of higher frequencies works with a fixed
# buffer size (see the WAX-ML module
# [`wax.modules.Buffer`](https://wax-ml.readthedocs.io/en/latest/_autosummary/wax.modules.buffer.html#module-wax.modules.buffer))
# which allows us to use JAX / XLA optimizations and have efficient processing.
#
# Let's illustrate with a small example how `wax.stream.Stream` synchronizes data streams.
#
# Let's use the dataset "air temperature" with :
# - An air temperature is defined with hourly resolution.
# - A "fake" ground temperature is defined with a daily resolution as the air temperature minus 10 degrees.

# + tags=[]
import xarray as xr

dataset = xr.tutorial.open_dataset("air_temperature")
dataset["ground"] = dataset.air.resample(time="d").last().rename({"time": "day"}) - 10
# -

# Let's see what this dataset looks like:

dataset

# + tags=[]
from wax.accessors import register_wax_accessors

register_wax_accessors()

# + tags=[]
from wax.modules import EWMA


def my_custom_function(dataset):
    return {
        "air_10": EWMA(1.0 / 10.0)(dataset["air"]),
        "air_100": EWMA(1.0 / 100.0)(dataset["air"]),
        "ground_100": EWMA(1.0 / 100.0)(dataset["ground"]),
    }


# -

results, state = dataset.wax.stream(
    local_time="time", ffills={"day": 1}, pbar=True
).apply(my_custom_function, format_dims=dataset.air.dims)

_ = results.isel(lat=0, lon=0).drop(["lat", "lon"]).to_pandas().plot(figsize=(12, 8))



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Universal Exponential moving average module and unroll implementations.
"""
from collections import namedtuple
from typing import Any, Tuple, cast

import haiku as hk
import jax
import jax.numpy as jnp
import tensorflow as tf
from jax.lax import fori_loop
from jax.tree_util import tree_flatten, tree_leaves, tree_map, tree_unflatten
from tqdm.auto import tqdm

import wax.external.eagerpy as ep
from wax.external.eagerpy import eager_function


class EagerEWMA(hk.Module):
    """Universal Exponential moving average module."""

    def __init__(self, alpha, adjust=True, name=None):
        super().__init__(name=name)
        self.alpha = alpha
        self.adjust = adjust

    @eager_function
    def __call__(self, x):

        mean = hk.get_state(
            "mean",
            shape=x.shape,
            dtype=x.dtype,
            init=lambda *_: ep.full_like(x, ep.nan),
        )
        count = hk.get_state(
            "count",
            shape=x.shape,
            dtype=x.dtype,
            init=lambda *_: ep.full_like(x, 0.0),
        )

        alpha = hk.get_parameter(
            "alpha",
            shape=(1,),
            dtype=x.dtype,
            init=lambda *_: type(x)(self.alpha).astype(dtype=x.dtype).reshape(1),
        )

        mean = ep.where(ep.isnan(mean), x, mean)
        count = ep.where(ep.logical_not(ep.isnan(x)), count + 1, count)

        if self.adjust == "linear":
            tscale = 1.0 / alpha
            tscale = ep.where(count < tscale, count, tscale)
            alpha = ep.where(tscale > 0, 1.0 / tscale, ep.nan)

        elif self.adjust:
            # exponential scheme (as in pandas)
            alpha = alpha / (1.0 - (1.0 - alpha) ** (count))

        # update mean  if x is not nan
        mean = ep.where(
            ep.logical_not(ep.isnan(x)),
            alpha * x + (1.0 - alpha) * mean,
            mean,
        )

        hk.set_state("mean", mean)
        hk.set_state("count", count)
        return mean


@eager_function
def static_unroll_universal(
    fun: hk.TransformedWithState,
    params: Any,
    state: Any,
    rng: jnp.ndarray,
    skip_first: bool = False,
    *args,
    pbar=True,
    compile=None,
    **kwargs,
):
    """Unroll a TransformedWithState function using static_scan implemented
    in an universal way.

    Args:
        fun : pair of pure functions (init, apply).
        params: parameters for the function.
        state : state for the function.
        rng: random number generator key.
        skip_first : if True, first value of the sequence is not used in apply.
        args, kwargs : Nested datastructure with sequences as leaves passed to init and apply
            of the TransformedWithState pair.
        pbar : if true, activate progress bar.
        compile : if true, compile the the appropriate backend.
    """
    fun = type(fun)(eager_function(fun.init), eager_function(fun.apply))

    # init
    xs = (args, kwargs)
    x0 = tree_map(lambda x: x[0], xs)
    fun_init_params, fun_init_state = fun.init(rng, *x0[0], **x0[1])
    params = fun_init_params if params is None else params
    state = fun_init_state if state is None else state
    # apply
    if skip_first:
        xs = tree_map(lambda x: x[1:], xs)

    if compile == "jax":
        fun = type(fun)(fun.init, jax.jit(fun.apply))
    if compile == "tensorflow":
        fun = type(fun)(fun.init, tf.function(fun.apply))

    xs, treedef = tree_flatten(xs)
    T = len(xs[0])

    output_template, _ = fun.apply(params, state, rng, *x0[0], **x0[1])
    output_template, output_treedef = tree_flatten(output_template)
    output_sequence: Tuple = tuple(map(lambda x: [], output_template))

    steps = range(T)
    if pbar:
        steps = tqdm(list(steps))
    for i in steps:
        args_i, kwargs_i = tree_unflatten(
            treedef, tuple(map(lambda x: cast(Tuple, x)[i], xs))
        )
        outputs, state = fun.apply(params, state, rng, *args_i, **kwargs_i)
        outputs = tree_leaves(outputs)
        for os, o in zip(output_sequence, outputs):
            os.append(o)

    @eager_function
    def _stack(x):
        return ep.stack(x)

    output_sequence = tuple(map(lambda x: _stack(x), output_sequence))
    output_sequence = tree_unflatten(output_treedef, output_sequence)
    return output_sequence, state


@eager_function
def dynamic_unroll_fori_loop_universal(
    fun: hk.TransformedWithState,
    params: Any,
    state: Any,
    rng: jnp.ndarray,
    skip_first: bool = False,
    *args,
    **kwargs,
):
    """Unroll a TransformedWithState function using fori_loop in an universal way.

    Args:
        fun : pair of pure functions (init, apply).
        params: parameters for the function.
        state : state for the function.
        rng: random number generator key.
        skip_first : if true, first value of the sequence is not used in apply.
        args, kwargs : Nested datastructure with sequences as leaves passed to init and apply
            of the TransformedWithState pair.
    """

    LoopState = namedtuple("LoopState", "params, state, rng, x, output_sequence")
    fun = type(fun)(eager_function(fun.init), eager_function(fun.apply))

    # init
    xs = (args, kwargs)
    x0 = tree_map(lambda x: x[0], xs)
    fun_init_params, fun_init_state = fun.init(rng, *x0[0], **x0[1])
    params = fun_init_params if params is None else params
    state = fun_init_state if state is None else state
    # apply
    if skip_first:
        xs = tree_map(lambda x: x[1:], xs)

    @jax.jit
    def body_fun(i, loop_state):
        params, prev_state, rng, inputs, output_sequence = loop_state
        args_i, kwargs_i = tree_map(lambda x: x[i], inputs)
        outputs, next_state = eager_function(fun.apply)(
            params, prev_state, rng, *args_i, **kwargs_i
        )
        output_sequence = ep.index_update(output_sequence, i, outputs)
        return LoopState(
            params,
            next_state,
            rng,
            inputs,
            output_sequence,
        )

    output_template, _ = fun.apply(params, state, rng, *x0[0], **x0[1])
    T = len(tree_flatten(xs)[0][0])

    @eager_function
    def _full_like_T_nan(x):
        x = ep.stack([x] * T)
        return ep.full_like(x, ep.nan)

    output_sequence = tree_map(lambda x: _full_like_T_nan(x), output_template)
    loop_state = LoopState(params, state, rng, xs, output_sequence)

    # run loop
    loop_state = fori_loop(0, T, body_fun, loop_state)

    # format results
    final_state = loop_state.state
    output_sequence = loop_state.output_sequence
    return output_sequence, final_state


@eager_function
def dynamic_unroll_universal(
    fun: hk.TransformedWithState,
    params: Any,
    state: Any,
    rng: jnp.ndarray,
    skip_first: bool = False,
    *args,
    **kwargs,
):
    """Unroll a TransformedWithState function using jax.lax.scan.

    Args:
        fun : pair of pure functions (init, apply).
        params: parameters for the function.
        state : state for the function.
        rng: random number generator key.
        skip_first : if true, first value of the sequence is not used in apply.
        args, kwargs : Nested datastructure with sequences as leaves passed to init and apply
            of the TransformedWithState pair.

    We can use scan because the output mean is in the state!

    See : https://jax.readthedocs.io/en/latest/_autosummary/jax.lax.scan.html
    """
    fun = type(fun)(eager_function(fun.init), eager_function(fun.apply))

    xs = (args, kwargs)
    x0 = tree_map(lambda x: x[0], xs)
    fun_init_params, fun_init_state = fun.init(rng, *x0[0], **x0[1])
    params = fun_init_params if params is None else params
    state = fun_init_state if state is None else state
    if skip_first:
        xs = tree_map(lambda x: x[1:], xs)

    def scan_f(prev_state, inputs):
        outputs, next_state = fun.apply(
            params, prev_state, rng, *inputs[0], **inputs[1]
        )
        return next_state, outputs

    final_state, output_sequence = jax.lax.scan(scan_f, init=state, xs=xs)
    return output_sequence, final_state


@tf.function
@tf.autograph.experimental.do_not_convert
def dynamic_unroll_tf(fun, params, state, rng, skip_first, x):
    """Unroll a TransformedWithState function using tensorflow
    scan function.

    Args:
        fun : pair of pure functions (init, apply).
        params: parameters for the function.
        state : state for the function.
        rng: random number generator key.
        skip_first : if true, first value of the sequence is not used in apply.
        x : Nested datastructure with sequences as leaves passed to init and apply
            of the TransformedWithState pair.
    """
    rng = rng if rng is not None else next(hk.PRNGSequence(42))

    def scan_f(prev_state, inputs):
        outputs, next_state = eager_function(fun.apply)(params, prev_state, rng, inputs)
        return next_state

    params, fun_init_state = eager_function(fun.init)(rng, x[0])
    state = fun_init_state if state is None else state

    assert not isinstance(x, tuple)
    state_sequence = tf.scan(scan_f, x, initializer=state)

    # TODO: find a more generic implementation
    output_sequence = state_sequence["eager_ewma"]["mean"]

    final_state = tree_map(lambda x: x[-1], state_sequence["eager_ewma"])
    return output_sequence, final_state



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Online supervised learner."""
from typing import Any, Callable, NamedTuple, Tuple

import haiku as hk
import jax
import jax.numpy as jnp
import optax

from wax.modules import FillNanInf


class ParamsState(NamedTuple):
    params: Any
    state: Any


class OnlineSupervisedLearnerInfo(NamedTuple):
    loss: float
    params: Any


class OnlineSupervisedLearner(hk.Module):
    """Online supervised learner."""

    def __init__(
        self,
        model: Any,
        opt: Any,
        loss: Callable,
        grads_fill_nan_inf=True,
        name: str = None,
    ):
        """Initialize module.

        Args:
            model : model to optimize.
            opt : optimizer.
            loss : loss function.
            name : name of the module
            grads_fill_nan_inf: if true, fill nan and
                +/- infinite values in gradients with zeros.
        """
        super().__init__(name=name)
        self.model = model
        self.opt = opt
        self.loss = loss
        self.grads_fill_nan_inf = grads_fill_nan_inf

    def __call__(
        self, x: jnp.ndarray, y: jnp.ndarray
    ) -> Tuple[jnp.ndarray, OnlineSupervisedLearnerInfo]:
        """Update learner.

        Args:
            x : features
            y: target
        """

        step = hk.get_state("step", [], init=lambda *_: 0)
        params, state = hk.get_state(
            "model_params_state",
            [],
            init=lambda *_: ParamsState(*self.model.init(hk.next_rng_key(), x)),
        )
        opt_state = hk.get_state("opt_state", [], init=lambda *_: self.opt.init(params))

        @jax.jit
        def _loss(params, state, x, y):
            y_pred, state = self.model.apply(params, state, None, x)
            return self.loss(y_pred, y)

        # compute loss and gradients
        l, grads = jax.value_and_grad(_loss)(params, state, x, y)

        if self.grads_fill_nan_inf:
            grads = FillNanInf()(grads)

        # compute prediction and update model state
        y_pred, state = self.model.apply(params, state, None, x)

        # update optimizer state
        grads, opt_state = self.opt.update(grads, opt_state)

        # update params
        params = optax.apply_updates(params, grads)

        step += 1
        hk.set_state("step", step)
        hk.set_state("model_params_state", ParamsState(params, state))
        hk.set_state("opt_state", opt_state)

        info = OnlineSupervisedLearnerInfo(loss=l, params=params)
        return y_pred, info



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Compute exponentially weighted variance."""
import haiku as hk
import jax.numpy as jnp


class EWMVar(hk.Module):
    """Compute exponentially weighted variance.

    To calculate the variance we use the fact that Var(X) = Mean(x^2) - Mean(x)^2 and internally
    we use the exponentially weighted mean of x/x^2 to calculate this.

    References
    ----------
    [^1]: [Finch, T., 2009. Incremental calculation of weighted mean and variance. University of Cambridge, 4(11-5), pp.41-42.](https://fanf2.user.srcf.net/hermes/doc/antiforgery/stats.pdf) # noqa
    """

    def __init__(self, alpha: float = 0.5, adjust: bool = True, name: str = None):
        """Initialize module.

        Args:
            alpha: alpha parameter of the exponential moving average.
            adjust: if true, implement a non-stationary filter with exponential initialization
                scheme. If "linear", implement a non-stationary filter with linear initialization.
            name : name of the module instance.
        """
        super().__init__(name=name)
        self.alpha = alpha
        self.adjust = adjust

    def __call__(self, x):
        """Implement last formula in [1]:
        .. maths::
            diff := x - mean
            incr := alpha * diff
            mean := mean + incr
            variance := (1 - alpha) * (variance + diff * incr)
        """
        mean = hk.get_state(
            "mean",
            shape=x.shape,
            dtype=x.dtype,
            init=lambda shape, dtype: jnp.full(shape, jnp.nan, dtype),
        )
        variance = hk.get_state(
            "variance",
            shape=x.shape,
            dtype=x.dtype,
            init=lambda shape, dtype: jnp.full(shape, jnp.nan, dtype),
        )
        count = hk.get_state(
            "count",
            shape=x.shape,
            dtype=x.dtype,
            init=lambda shape, dtype: jnp.full(shape, 0.0, dtype),
        )

        alpha = hk.get_parameter(
            "alpha",
            shape=[],
            dtype=x.dtype,
            init=lambda *_: jnp.array(self.alpha),
        )

        # initialization on first non-nan value
        mean = jnp.where(jnp.isnan(mean), x, mean)
        variance = jnp.where(jnp.isnan(variance), 0.0, variance)

        mask = jnp.logical_not(jnp.isnan(x))
        count = jnp.where(mask, count + 1, count)

        # alpha adjustement scheme
        if self.adjust == "linear":
            tscale = 1.0 / alpha
            tscale = jnp.where(count < tscale, count, tscale)
            alpha = jnp.where(tscale > 0, 1.0 / tscale, jnp.nan)
        elif self.adjust:
            # exponential scheme (as in pandas)
            alpha = alpha / (1.0 - (1.0 - alpha) ** (count))

        diff = x - mean
        incr = alpha * diff

        # update state
        mean = jnp.where(mask, mean + incr, mean)
        variance = jnp.where(mask, (1 - alpha) * (variance + diff * incr), variance)

        hk.set_state("mean", mean)
        hk.set_state("variance", variance)
        hk.set_state("count", count)

        return variance



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ARMA linear filter."""
import haiku as hk
import jax.numpy as jnp

from wax.modules.buffer import Buffer
from wax.modules.fill_nan_inf import FillNanInf


class ARMA(hk.Module):
    """ARMA linear filter."""

    def __init__(self, alpha, beta, name=None):
        super().__init__(name=name)
        self.alpha = alpha
        self.beta = beta

    def __call__(self, eps):
        eps_buffer = Buffer(len(self.beta) + 1, 0.0, name="eps_buffer")(eps)[:-1]

        y_buffer = hk.get_state(
            "y_buffer",
            eps.shape,
            eps.dtype,
            init=lambda shape, dtype: jnp.zeros(
                ((len(self.alpha),) + shape), dtype=dtype
            ),
        )

        y = self.alpha @ y_buffer + self.beta @ eps_buffer + eps

        y = FillNanInf()(y)

        # reshape with shape of eps (for scalar case)
        y = y.reshape(eps.shape)

        y_buffer = Buffer(len(self.alpha), 0, name="y_buffer")(y)

        hk.set_state("y_buffer", y_buffer)
        return y



# -*- coding: utf-8 -*-
# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:light,md
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.5'
#       jupytext_version: 1.13.0
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# # 🔄 Online learning for time series prediction 🔄
#
#
# In [1], the authors develop an online learning method to predict time-series
# generated by and ARMA (autoregressive moving average) model.
#
# They develop an effective online learning algorithm based on an **improper learning** approach which consists to use an AR model for prediction with sufficiently long horizon, together with an online update of the prediction model parameters using either an "online Newton" algorithm [2] or a stochastic gradient descent algorithm.
#
# This effective approach adresses the prediction problem, without assuming that the noise terms are Gaussian, identically distributed or even independent.
# Furthermore, they show that their algorithm’s performances asymptotically approaches the performance of the best ARMA model in hindsight.
#
# We use WAX-ML to reproduce their empirical results.
#
# We first focus in the reproduction of the "setting 1" for sanity checks and show how to setup a training environment with WAX-ML to study improper learning of ARMA time-series models.
#
# We then study the behavior the method with different optimizers in the non-stationary environements proposed in [1] (settings 2, 3, and 4).
#
# We use the following modules from WAX-ML:
#   * ARMA : to generate a modeled time-series
#   * SNARIMAX : to adaptively learn to predict the generated time-series.
#   * GymFeedback: To setup a training loop.
#   * VMap: to add batch dimensions to the training loop
#   * optim.newton: a newton algorithm as used in [1] and developped in [2]. It extends `optax` optimizers.
#   * OnlineOptimizer: A wrapper for a model with loss and and optimizer for online learning.
#
# ## References
#
# [1] [Anava, O., Hazan, E., Mannor, S. and Shamir, O., 2013, June. Online learning for time series prediction. In Conference on learning theory (pp. 172-184)](https://arxiv.org/pdf/1302.6927.pdf)
#
# [2] [Hazan, E., Agarwal, A. and Kale, S., 2007. Logarithmic regret algorithms for
# online convex optimization. Machine Learning, 69(2-3), pp.169-192](https://link.springer.com/content/pdf/10.1007/s10994-007-5016-8.pdf)

# %pylab inline
# %load_ext autoreload
# %autoreload 2


# +
from typing import Any, NamedTuple

import haiku as hk
import jax
import jax.numpy as jnp
import numpy as onp
import optax
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt
from tqdm.auto import tqdm

from wax.modules import (
    ARMA,
    SNARIMAX,
    GymFeedback,
    Lag,
    OnlineOptimizer,
    UpdateParams,
    VMap,
)
from wax.optim import newton
from wax.unroll import unroll_transform_with_state

# + tags=["parameters"]
T = 10000
N_BATCH = 20
N_STEP_SIZE = 10
N_EPS = 5
# -

# ## ARMA
#

# Let's generate a sample of the "setting 1" of [1]:

# +
alpha = jnp.array([0.6, -0.5, 0.4, -0.4, 0.3])
beta = jnp.array([0.3, -0.2])

rng = jax.random.PRNGKey(42)
eps = jax.random.normal(rng, (1000,))
sim = unroll_transform_with_state(lambda eps: ARMA(alpha, beta)(eps))
params, state = sim.init(rng, eps)
y, state = sim.apply(params, state, rng, eps)
pd.Series(y).plot()


# -

# ## SNARIMAX
#

# Let's setup an online model to try to learn the dynamic of the time-series.

# First let's run the filter with it's initial random weights.

# +
def predict(y, X=None):
    return SNARIMAX(10, 0, 0)(y, X)


sim = unroll_transform_with_state(predict)
rng = jax.random.PRNGKey(42)
eps = jax.random.normal(rng, (10,))
params, state = sim.init(rng, y)
(y_pred, _), state = sim.apply(params, state, rng, y)

pd.Series((y - y_pred)).plot()


# +
def evaluate(y_pred, y):
    return jnp.linalg.norm(y_pred - y) ** 2, {}


def lag(shift=1):
    def __call__(y, X=None):
        yp = Lag(shift)(y)
        Xp = Lag(shift)(X) if X is not None else None
        return yp, Xp

    return __call__


def predict_and_evaluate(y, X=None):
    # predict with lagged data
    y_pred, pred_info = predict(*lag(1)(y, X))

    # evaluate loss with actual data
    loss, loss_info = evaluate(y_pred, y)

    return loss, dict(pred_info=pred_info, loss_info=loss_info)


# +
sim = unroll_transform_with_state(predict_and_evaluate)
rng = jax.random.PRNGKey(42)
eps = jax.random.normal(rng, (1000,))
params, state = sim.init(rng, y)
(loss, _), state = sim.apply(params, state, rng, y)

pd.Series(loss).expanding().mean().plot()
# -

# Since the model is not trained and the coefficient of the SNARIMAX filter a choosen randomly, the loss may diverge

params


# ## Learn

# To learn the model parameters we will use the `OnlineOptimizer` of WAX-ML.

# ### Setup projection

# We can setup a projection for the parameters:
#
#


def project_params(params, opt_state=None):
    w = params["snarimax/~/linear"]["w"]
    w = jnp.clip(w, -1, 1)
    params["snarimax/~/linear"]["w"] = w
    return params


project_params(hk.data_structures.to_mutable_dict(params))


def learn(y, X=None):
    optim_res = OnlineOptimizer(
        predict_and_evaluate, optax.sgd(1.0e-2), project_params=project_params
    )(y, X)
    return optim_res


# Let's train:

# +
sim = unroll_transform_with_state(learn)
rng = jax.random.PRNGKey(42)
eps = jax.random.normal(rng, (1000,))
params, state = sim.init(rng, eps)
optim_res, state = sim.apply(params, state, rng, eps)

pd.Series(optim_res.loss).expanding().mean().plot()
# -

# Let's look at the latest weights:

jax.tree_map(lambda x: x[-1], optim_res.updated_params)


# ## Learn and Forecast
#
#
#


class ForecastInfo(NamedTuple):
    optim: Any
    forecast: Any


def learn_and_forecast(y, X=None):
    optim_res = OnlineOptimizer(
        predict_and_evaluate, optax.sgd(1.0e-3), project_params=project_params
    )(*lag(1)(y, X))

    predict_params = optim_res.updated_params

    forecast, forecast_info = UpdateParams(predict)(predict_params, y, X)
    return forecast, ForecastInfo(optim_res, forecast_info)


# +
sim = unroll_transform_with_state(learn_and_forecast)
rng = jax.random.PRNGKey(42)
eps = jax.random.normal(rng, (1000,))
params, state = sim.init(rng, y)
(forecast, info), state = sim.apply(params, state, rng, y)

pd.Series(info.optim.loss).expanding().mean().plot()


# -

# ## Gym simulation

# Now let's wrapup the training loop in a gym feedback loop.
#
#

# ### Environment
#

# Let's build an environment corresponding to "setting 1" in [1]
#
#


def build_env():
    def env(action, obs):
        y_pred, eps = action, obs
        ar_coefs = jnp.array([0.6, -0.5, 0.4, -0.4, 0.3])
        ma_coefs = jnp.array([0.3, -0.2])

        y = ARMA(ar_coefs, ma_coefs)(eps)
        # prediction used on a fresh y observation.
        rw = -((y - y_pred) ** 2)

        env_info = {"y": y, "y_pred": y_pred}
        obs = y
        return rw, obs, env_info

    return env


env = build_env()


# ### Agent

# Let's build an agent:

from optax._src.base import OptState

# + tags=[]


def build_agent(time_series_model=None, opt=None):
    if time_series_model is None:
        time_series_model = lambda y, X: SNARIMAX(10)(y, X)

    if opt is None:
        opt = optax.sgd(1.0e-3)

    class AgentInfo(NamedTuple):
        optim: Any
        forecast: Any

    class ModelWithLossInfo(NamedTuple):
        pred: Any
        loss: Any

    def agent(obs):
        if isinstance(obs, tuple):
            y, X = obs
        else:
            y = obs
            X = None

        def evaluate(y_pred, y):
            return jnp.linalg.norm(y_pred - y) ** 2, {}

        def model_with_loss(y, X=None):
            # predict with lagged data
            y_pred, pred_info = time_series_model(*lag(1)(y, X))

            # evaluate loss with actual data
            loss, loss_info = evaluate(y_pred, y)

            return loss, ModelWithLossInfo(pred_info, loss_info)

        def project_params(params: Any, opt_state: OptState = None):
            del opt_state
            return jax.tree_map(lambda w: jnp.clip(w, -1, 1), params)

        def split_params(params):
            def filter_params(m, n, p):
                # print(m, n, p)
                return m.endswith("snarimax/~/linear") and n == "w"

            return hk.data_structures.partition(filter_params, params)

        def learn_and_forecast(y, X=None):
            optim_res = OnlineOptimizer(
                model_with_loss,
                opt,
                project_params=project_params,
                split_params=split_params,
            )(*lag(1)(y, X))

            predict_params = optim_res.updated_params

            y_pred, forecast_info = UpdateParams(time_series_model)(
                predict_params, y, X
            )
            return y_pred, AgentInfo(optim_res, forecast_info)

        return learn_and_forecast(y, X)

    return agent


# -


agent = build_agent()


# ### Gym loop
#
#
#


def gym_loop(eps):
    return GymFeedback(agent, env)(eps)


rng = jax.random.PRNGKey(42)
eps = jax.random.normal(rng, (T,))
sim = unroll_transform_with_state(gym_loop)
params, state = sim.init(rng, eps)
(gym, info), state = sim.apply(params, state, rng, eps)
pd.Series(info.agent.optim.loss).expanding().mean().plot(label="agent loss")
pd.Series(-gym.reward).expanding().mean().plot(xlim=(0, 100), label="env loss")
plt.legend()

pd.Series(-gym.reward).expanding().mean().plot()  # ylim=(0.09, 0.15))


# We see that the agent suffers the same loss as the environment but with a time lag.

# ## Batch simulations

# ### Average over 20 experiments

# ### Slow version
#
# First, let's do it "naively" by doing a simple python "for loop".

# +
# %%time
rng = jax.random.PRNGKey(42)
sim = unroll_transform_with_state(gym_loop)

res = {}
for i in tqdm(onp.arange(N_BATCH)):
    rng, _ = jax.random.split(rng)
    eps = jax.random.normal(rng, (T,)) * 0.3
    params, state = sim.init(rng, eps)
    (gym_output, gym_info), final_state = sim.apply(params, state, rng, eps)
    res[i] = gym_info
# -

pd.DataFrame({k: pd.Series(v.agent.optim.loss) for k, v in res.items()}).mean(
    1
).expanding().mean().plot(ylim=(0.09, 0.15))

# ### Fast version with vmap
#
# Instead of using a "for loop" we can use jax's vmap transformation function!
#

# +
# %%time
rng = jax.random.PRNGKey(42)
eps = jax.random.normal(rng, (N_BATCH, T)) * 0.3

rng = jax.random.PRNGKey(42)
rng = jax.random.split(rng, num=N_BATCH)
sim = unroll_transform_with_state(gym_loop)
params, state = jax.vmap(sim.init)(rng, eps)
(gym_output, gym_info), final_state = jax.vmap(sim.apply)(params, state, rng, eps)
# -

# This is much faster!

pd.DataFrame(gym_info.agent.optim.loss).mean().expanding().mean().plot(
    ylim=(0.09, 0.15)
)

# +
i_batch = 0
w = gym_info.agent.optim.updated_params["snarimax/~/linear"]["w"][i_batch, :, :, 0]
w = pd.DataFrame(w)

ax = w.plot(title=f"weights on batch {i_batch}")
ax.legend(bbox_to_anchor=(1.0, 1.0))
ax.set_xlabel("time")
ax.set_ylabel("weight value")


plt.figure()
w.iloc[-1][::-1].plot(kind="bar")

# +
w = gym_info.agent.optim.updated_params["snarimax/~/linear"]["w"].mean(axis=0)[:, :, 0]
w = pd.DataFrame(w)
ax = w.plot(title="averaged weights (over batches)")
ax.legend(bbox_to_anchor=(1.0, 1.0))
ax.set_xlabel("time")
ax.set_ylabel("weight")

plt.figure()
w.iloc[-1][::-1].plot(kind="bar")
# -

# ### With VMap module
#
# We can use the wrapper module `VMap` of WAX-ML. It permits to have an ever simpler syntax.
#

# > Note: we have to swap the position of time and batch dimensions in the generation of the noise variable `eps`.

# +
# %%time
rng = jax.random.PRNGKey(42)
eps = jax.random.normal(rng, (T, N_BATCH)) * 0.3


def batched_gym_loop(eps):
    return VMap(gym_loop)(eps)


sim = unroll_transform_with_state(batched_gym_loop)

rng = jax.random.PRNGKey(43)
params, state = sim.init(rng, eps)
(gym_output, gym_info), final_state = sim.apply(params, state, rng, eps)
# -

pd.DataFrame(gym_info.agent.optim.loss).shape

pd.DataFrame(gym_info.agent.optim.loss).mean(1).expanding().mean().plot(
    ylim=(0.09, 0.15)
)

# +
i_batch = 0
w = gym_info.agent.optim.updated_params["snarimax/~/linear"]["w"][:, i_batch, :, 0]
w = pd.DataFrame(w)

ax = w.plot(title=f"weights on batch {i_batch}")
ax.legend(bbox_to_anchor=(1.0, 1.0))
ax.set_xlabel("time")
ax.set_ylabel("weight value")


plt.figure()
w.iloc[-1][::-1].plot(kind="bar")

# +
w = gym_info.agent.optim.updated_params["snarimax/~/linear"]["w"].mean(axis=1)[:, :, 0]
w = pd.DataFrame(w)
ax = w.plot(title="averaged weights (over batches)")
ax.legend(bbox_to_anchor=(1.0, 1.0))
ax.set_xlabel("time")
ax.set_ylabel("weight")

plt.figure()
w.iloc[-1][::-1].plot(kind="bar")
# -

# ### Taking mean inside simulation

# +
# %%time


def add_batch(fun, take_mean=True):
    def fun_batch(*args, **kwargs):
        res = VMap(fun)(*args, **kwargs)
        if take_mean:
            res = jax.tree_map(lambda x: x.mean(axis=0), res)
        return res

    return fun_batch


gym_loop_batch = add_batch(gym_loop)
sim = unroll_transform_with_state(gym_loop_batch)

rng = jax.random.PRNGKey(42)
eps = jax.random.normal(rng, (T, N_BATCH)) * 0.3

params, state = sim.init(rng, eps)
(gym_output, gym_info), final_state = sim.apply(params, state, rng, eps)
# -

pd.Series(-gym_output.reward).expanding().mean().plot(ylim=(0.09, 0.15))

# +
w = gym_info.agent.optim.updated_params["snarimax/~/linear"]["w"][:, :, 0]
w = pd.DataFrame(w)
ax = w.plot(title="averaged weights (over batches)")
ax.legend(bbox_to_anchor=(1.0, 1.0))
ax.set_xlabel("time")
ax.set_ylabel("weight")

plt.figure()
w.iloc[-1][::-1].plot(kind="bar")
# -

# ## Hyper parameter tuning

# ### First order optimizers
#
# We will consider different first order optimizers, namely:
# - SGD
# - ADAM
# - ADAGRAD
#
# For each of them, we will scan the "step_size" parameter $\eta$.
#
# We will average results over batches of size 40.
#
# We will consider trajectories of size 10.000.
#
# Finally, we will pickup the best parameter based of the minimum averaged loss for the last 5000 time steps.

# +
# %%time

STEP_SIZE_idx = pd.Index(onp.logspace(-4, 1, 30), name="step_size")
STEP_SIZE = jax.device_put(STEP_SIZE_idx.values)
OPTIMIZERS = [optax.sgd, optax.adagrad, optax.rmsprop, optax.adam]

res = {}
for optimizer in tqdm(OPTIMIZERS):

    def gym_loop_scan_hparams(eps):
        def scan_params(step_size):
            return GymFeedback(build_agent(opt=optimizer(step_size)), env)(eps)

        res = VMap(scan_params)(STEP_SIZE)
        return res

    sim = unroll_transform_with_state(add_batch(gym_loop_scan_hparams))
    rng = jax.random.PRNGKey(42)
    eps = jax.random.normal(rng, (T, N_BATCH)) * 0.3

    params, state = sim.init(rng, eps)
    _res, state = sim.apply(params, state, rng, eps)
    res[optimizer.__name__] = _res

# -

ax = None
BEST_STEP_SIZE = {}
BEST_GYM = {}
for name, (gym, info) in res.items():

    loss = pd.DataFrame(-gym.reward, columns=STEP_SIZE).iloc[-5000:].mean()

    BEST_STEP_SIZE[name] = loss.idxmin()
    best_idx = loss.reset_index(drop=True).idxmin()
    BEST_GYM[name] = jax.tree_map(lambda x: x[:, best_idx], gym)

    ax = loss[loss < 0.15].plot(logx=True, logy=False, ax=ax, label=name)
plt.legend()

for name, gym in BEST_GYM.items():
    ax = (
        pd.Series(-gym.reward)
        .expanding()
        .mean()
        .plot(
            label=f"{name}    -    $\eta$={BEST_STEP_SIZE[name]:.2e}", ylim=(0.09, 0.15)
        )
    )
ax.legend(bbox_to_anchor=(1.0, 1.0))

pd.Series(BEST_STEP_SIZE).plot(kind="bar", logy=True)


# ### Newton algorithm

# Now let's consider the newton algorithm.

# First let's test it with one set of parameter with average over N_BATCH batches.

# +
# %%time


@add_batch
def gym_loop_newton(eps):
    return GymFeedback(build_agent(opt=newton(0.05, eps=20.0)), env)(eps)


sim = unroll_transform_with_state(gym_loop_newton)
rng = jax.random.PRNGKey(42)
eps = jax.random.normal(rng, (T, N_BATCH)) * 0.3

params, state = sim.init(rng, eps)
(gym, info), state = sim.apply(params, state, rng, eps)
# -


pd.Series(-gym.reward).expanding().mean().plot()

# +
# %%time

STEP_SIZE = pd.Index(onp.logspace(-2, 3, 10), name="step_size")
EPS = pd.Index(onp.logspace(-4, 3, 5), name="eps")

HPARAMS_idx = pd.MultiIndex.from_product([STEP_SIZE, EPS])
HPARAMS = jnp.stack(list(map(onp.array, HPARAMS_idx)))


@add_batch
def gym_loop_scan_hparams(eps):
    def scan_params(hparams):
        step_size, newton_eps = hparams
        agent = build_agent(opt=newton(step_size, eps=newton_eps))
        return GymFeedback(agent, env)(eps)

    return VMap(scan_params)(HPARAMS)


sim = unroll_transform_with_state(gym_loop_scan_hparams)
rng = jax.random.PRNGKey(42)
eps = jax.random.normal(rng, (T, N_BATCH)) * 0.3

params, state = sim.init(rng, eps)
res_newton, state = sim.apply(params, state, rng, eps)
# -


gym_newton, info_newton = res_newton

loss_newton = pd.DataFrame(-gym_newton.reward, columns=HPARAMS_idx).mean().unstack()


loss_newton = (
    pd.DataFrame(-gym_newton.reward, columns=HPARAMS_idx).iloc[-5000:].mean().unstack()
)


# +
sns.heatmap(loss_newton[loss_newton < 0.4], annot=True, cmap="YlGnBu")

STEP_SIZE, NEWTON_EPS = loss_newton.stack().idxmin()

x = -gym_newton.reward[-5000:].mean(axis=0)
x = jax.ops.index_update(x, jnp.isnan(x), jnp.inf)
I_BEST_PARAM = jnp.argmin(x)


BEST_NEWTON_GYM = jax.tree_map(lambda x: x[:, I_BEST_PARAM], gym_newton)
print("Best newton parameters: ", STEP_SIZE, NEWTON_EPS)

# +
for name, gym in BEST_GYM.items():
    pd.Series(-gym.reward).rolling(5000, min_periods=5000).mean().plot(
        label=f"{name}    -    $\eta$={BEST_STEP_SIZE[name]:.2e}", ylim=(0.09, 0.1)
    )

gym = BEST_NEWTON_GYM
ax = (
    pd.Series(-gym.reward)
    .rolling(5000, min_periods=5000)
    .mean()
    .plot(
        label=f"Newton    -    $\eta$={STEP_SIZE:.2e},    $\epsilon$={NEWTON_EPS:.2e}"
    )
)
ax.legend(bbox_to_anchor=(1.0, 1.0))
plt.title("Rolling mean of loss (5000) time-steps")
# -

for name, gym in BEST_GYM.items():
    pd.Series(-gym.reward).expanding().mean().plot(
        label=f"{name}    -    $\eta$={BEST_STEP_SIZE[name]:.2e}", ylim=(0.09, 0.15)
    )
gym = BEST_NEWTON_GYM
ax = (
    pd.Series(-gym.reward)
    .expanding()
    .mean()
    .plot(
        label=f"Newton    -    $\eta$={STEP_SIZE:.2e},    $\epsilon$={NEWTON_EPS:.2e}"
    )
)
ax.legend(bbox_to_anchor=(1.0, 1.0))


# In agreement with results in [1], we see that Newton's algorithm performs much better than SGD.
#
# In addition, we note that:
# - ADAGRAD performormance is between newton and sgd.
# - RMSPROPR and ADAM does not perform well in this online setting.



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import haiku as hk
import jax
import jax.numpy as jnp
import pytest

from wax.compile import jit_init_apply
from wax.modules.buffer import Buffer


@pytest.mark.parametrize("use_jit", [False, True])
def test_buffer(use_jit):
    seq = hk.PRNGSequence(42)
    x = jax.random.normal(next(seq), (2, 3))

    @hk.transform_with_state
    def buffer(x):
        return Buffer(2, return_state=False, name="buf")(x)

    if use_jit:
        buffer = jit_init_apply(buffer)

    params, state = buffer.init(next(seq), x)
    assert jnp.isnan(state["buf"]["buffer_state"].buffer).all()
    assert state["buf"]["buffer_state"].len_buffer == 0
    assert state["buf"]["buffer_state"].i_start == 2

    output, state = buffer.apply(params, state, next(seq), x)
    assert len(output) == 2
    if use_jit:
        assert state["buf"]["buffer_state"].buffer is not output
        assert jnp.allclose(state["buf"]["buffer_state"].buffer, output, equal_nan=True)
    else:
        assert state["buf"]["buffer_state"].buffer is output
    assert state["buf"]["buffer_state"].len_buffer == 1
    assert state["buf"]["buffer_state"].i_start == 1
    assert jnp.allclose(output[-1], x)
    assert jnp.isnan(output[-2]).all()
    x1 = x

    x = jax.random.normal(next(seq), (2, 3))
    output, state = buffer.apply(params, state, next(seq), x)
    assert len(output) == 2
    if use_jit:
        assert state["buf"]["buffer_state"].buffer is not output
        assert jnp.allclose(state["buf"]["buffer_state"].buffer, output, equal_nan=True)
    else:
        assert state["buf"]["buffer_state"].buffer is output
    assert state["buf"]["buffer_state"].len_buffer == 2
    assert state["buf"]["buffer_state"].i_start == 0
    assert jnp.allclose(output[-1], x)
    assert jnp.allclose(output[-2], x1)
    x2 = x

    x = jax.random.normal(next(seq), (2, 3))
    output, state = buffer.apply(params, state, next(seq), x)
    assert len(output) == 2
    if use_jit:
        assert state["buf"]["buffer_state"].buffer is not output
        assert jnp.allclose(state["buf"]["buffer_state"].buffer, output, equal_nan=True)
    else:
        assert state["buf"]["buffer_state"].buffer is output
    assert state["buf"]["buffer_state"].len_buffer == 2
    assert state["buf"]["buffer_state"].i_start == 0
    assert jnp.allclose(output[-1], x)
    assert jnp.allclose(output[-2], x2)



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest

import wax.external.eagerpy as ep


@pytest.mark.parametrize("axis", [0, 1, -1])
def test_kl_div_with_logits(dummy: ep.Tensor, axis: int) -> None:
    logits_p = logits_q = ep.arange(dummy, 12).float32().reshape((3, 4))
    assert (ep.kl_div_with_logits(logits_p, logits_q, axis=axis) == 0).all()



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any

import numpy as np
import pytest

import wax.external.eagerpy as ep


@pytest.mark.parametrize("fill_value", [0.0, 0, "0"])
def test_astensors_list_float(fill_value: Any) -> None:
    x = [
        fill_value,
    ] * 3
    ex = ep.astensors(x)
    assert isinstance(ex[0], ep.Tensor)
    ex_stacked = ep.stack(ex)
    x_stacked = ex_stacked.raw
    assert isinstance(x_stacked, np.ndarray)


def test_np_full_like() -> None:
    x, alpha = ep.astensors(
        np.array([1.10532079, 0.79016002, -1.45496991]), np.array(0.1)
    )
    x_like_alpha = ep.full_like(x, alpha)
    assert (x_like_alpha.raw == alpha.raw).all()



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Gym agent defined from an Haiku module"""

from dataclasses import dataclass, field
from typing import Any

import haiku as hk

from wax.gym.agent import Agent


@dataclass
class HaikuAgent(Agent):
    """Wrap a pair of init/apply function as a standard python class with state management."""

    agent: hk.TransformedWithState
    name: Any = None
    _seq: hk.PRNGSequence = field(
        init=False, default_factory=lambda: hk.PRNGSequence(42)
    )
    _params: Any = field(init=False, default=None)
    _state: Any = field(init=False, default=None)
    _initialized: bool = field(init=False, default=False)

    def _initialize(self, obs):
        self._params, self._state = self.agent.init(next(self._seq), obs)
        self._initialized = True

    def reset(self):
        self._params = None
        self._state = None
        self._initialized = False

    def __call__(self, obs):

        if not self._initialized:
            self._initialize(obs)

        action, self._state = self.agent.apply(
            self._params, self._state, next(self._seq), obs
        )
        return action



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from os.path import dirname as _dirname
from os.path import join as _join

from . import norms  # noqa: F401,E402
from . import types  # noqa: F401,E402
from . import utils  # noqa: F401,E402
from ._index import index  # noqa: F401,E402
from .astensor import as_raw_tensor  # noqa: F401,E402
from .astensor import as_raw_tensors  # noqa: F401,E402
from .astensor import astensor  # noqa: F401,E402
from .astensor import astensor_  # noqa: F401,E402
from .astensor import astensors  # noqa: F401,E402
from .astensor import astensors_  # noqa: F401,E402
from .astensor import eager_function  # noqa: F401,E402
from .convert import convert_to_tensor  # noqa: F401,E402
from .convert import convert_to_tensors  # noqa: F401,E402
from .framework import *  # noqa: F401,E402,F403
from .lib import *  # noqa: F401,E402,F403
from .modules import jax  # noqa: F401,E402
from .modules import numpy  # noqa: F401,E402
from .modules import tensorflow  # noqa: F401,E402
from .modules import torch  # noqa: F401,E402
from .tensor import JAXTensor  # noqa: F401,E402
from .tensor import NumPyTensor  # noqa: F401,E402
from .tensor import PyTorchTensor  # noqa: F401,E402
from .tensor import Tensor  # noqa: F401,E402
from .tensor import TensorFlowTensor  # noqa: F401,E402
from .tensor import TensorType  # noqa: F401,E402
from .tensor import istensor  # noqa: F401,E402

with open(_join(_dirname(__file__), "VERSION")) as _f:
    __version__ = _f.read().strip()



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Gym environment defined from an Haiku module"""

from dataclasses import dataclass, field
from typing import Any, Dict, Generator

import haiku as hk

from wax.gym.env import Env


@dataclass
class HaikuEnv(Env):
    """Wrap a pair of init/apply function as a standard python class
    with state management and Gym env API.

    Args:
        data_generator :
    """

    data_generator: Generator
    env: hk.TransformedWithState
    name: Any = None

    _info: Dict = field(init=False, default_factory=dict)
    _seq: hk.PRNGSequence = field(
        init=False, default_factory=lambda: hk.PRNGSequence(42)
    )
    _params: Any = field(init=False, default=None)
    _state: Any = field(init=False, default=None)
    _initialized: bool = field(init=False, default=False)

    def reset(self):
        raw_obs, self._info = next(self.data_generator)
        self._initialized = False
        return raw_obs

    def step(self, action):
        try:
            raw_obs, self._info = next(self.data_generator)
            done = False
        except StopIteration:
            obs = None
            rw = 0
            done = True
            return obs, rw, done, self._info

        if not self._initialized:
            self._params, self._state = self.env.init(next(self._seq), action, raw_obs)
            self._initialized = True

        (rw, obs), self._state = self.env.apply(
            self._params, self._state, next(self._seq), action, raw_obs
        )

        return obs, rw, done, self._info

    def __len__(self):
        return self._info.get("maxlen") if isinstance(self._info, dict) else None



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""We implement an online learning non-stationary linear regression problem.

We go there progressively by showing how a linear regression problem can be cast
into an online learning problem thanks to the `OnlineSupervisedLearner` module.

Then, in order to tackle a non-stationary linear regression problem (i.e. with a weight that can vary in time)
we reformulate the problem into a reinforcement learning problem that we implement with the `GymFeedBack` module of WAX-ML.

We then need to define an "agent" and an "environment" using simple functions or modules:
- The agent is responsible for learning the weights of its internal linear model.
- The environment is responsible for generating labels and evaluating the agent's reward metric.

We experiment with a non-stationary environment that returns the sign of the linear regression parameters at a given time step,
known only to the environment.

We will see that doing this is very simple with the WAX-ML tools and that the functional workflow it adopts
allows, each time we increase in complexity, to reuse the previously implemented transformations.


In this journey, we will use:
- Haiku basic linear module `hk.Linear`.
- Optax stochastic gradient descent optimizer: `sgd`.
- WAX-ML modules: `OnlineSupervisedLearner`, `Lag`, `GymFeedBack`
- WAX-ML helper functions: `dynamic_unroll`, `jit_init_apply`
"""
import haiku as hk
import jax
import jax.numpy as jnp
import optax
from jax.tree_util import tree_map

from wax.compile import jit_init_apply
from wax.modules import GymFeedback, Lag, OnlineSupervisedLearner
from wax.unroll import unroll


@jit_init_apply
@hk.transform_with_state
def linear_model(x):
    return hk.Linear(output_size=1, with_bias=False)(x)


def test_static_model():
    # First let's implement a simple linear regression
    # Let's generate a batch of data:

    seq = hk.PRNGSequence(42)
    T = 100
    N = 3
    X = jax.random.normal(next(seq), (T, N))
    w_true = jnp.ones(N)

    params, state = linear_model.init(next(seq), X[0])
    linear_model.apply(params, state, None, X[0])

    Y_pred = unroll(linear_model, rng=next(seq))(X)

    assert Y_pred.shape == (T, 1)

    noise = jax.random.normal(next(seq), (T,))
    Y = X.dot(w_true) + noise
    mean_loss = ((Y - Y_pred) ** 2).sum(axis=1).mean()
    assert mean_loss > 0


def generate_many_observations(T=300, sigma=1.0e-2, rng=None):
    rng = jax.random.PRNGKey(42) if rng is None else rng
    X = jax.random.normal(rng, (T, 3))
    noise = sigma * jax.random.normal(rng, (T,))
    w_true = jnp.ones(3)
    noise = sigma * jax.random.normal(rng, (T,))
    Y = X.dot(w_true) + noise
    return (X, Y)


def test_online_model():
    # # Online model

    opt = optax.sgd(1e-3)

    @jax.jit
    def loss(y_pred, y):
        return jnp.mean(jnp.square(y_pred - y))

    @jit_init_apply
    @hk.transform_with_state
    def learner(x, y):
        return OnlineSupervisedLearner(linear_model, opt, loss)(x, y)

    seq = hk.PRNGSequence(42)

    # generate data
    T = 300
    X, Y = generate_many_observations(T)

    # dynamic unroll the learner
    x0, y0 = tree_map(lambda x: x[0], (X, Y))
    (output, info) = unroll(learner, rng=next(seq))(X, Y)
    assert len(info.loss) == T
    assert len(info.params["linear"]["w"])


def linear_regression_agent(obs):
    x, y = obs

    opt = optax.sgd(1e-3)

    @jax.jit
    def loss(y_pred, y):
        return jnp.mean(jnp.square(y_pred - y))

    def learner(x, y):
        return OnlineSupervisedLearner(linear_model, opt, loss)(x, y)

    return learner(x, y)


def stationary_linear_regression_env(y_pred, raw_obs):
    # Only the environment now the true value of the parameters
    w_true = -jnp.ones(3)

    # The environment has its proper loss definition
    @jax.jit
    def loss(y_pred, y):
        return jnp.mean(jnp.square(y_pred - y))

    # raw observation contains features and generative noise
    x, noise = raw_obs

    # generate targets
    y = x @ w_true + noise
    obs = (x, y)

    y_previous = Lag(1)(y)
    # evaluate the prediction made by the agent
    reward = loss(y_pred, y_previous)

    info = {}
    return reward, obs, info


def generate_many_raw_observations(T=300, sigma=1.0e-2, rng=None):
    rng = jax.random.PRNGKey(42) if rng is None else rng
    X = jax.random.normal(rng, (T, 3))
    noise = sigma * jax.random.normal(rng, (T,))
    return (X, noise)


def test_online_recast_as_reinforcement_learning_pb():
    # # Online supervised learning recast as a reinforcement learning problem
    # obs = (x, y) are tuple observations.
    # raw_obs = (x, noise) consist in the feature and input noise.

    @hk.transform_with_state
    def gym_fun(raw_obs):
        return GymFeedback(
            linear_regression_agent,
            stationary_linear_regression_env,
            return_action=True,
        )(raw_obs)

    T = 300
    raw_observations = generate_many_raw_observations(T)
    rng = jax.random.PRNGKey(42)
    (gym_output, gym_info) = unroll(gym_fun, rng=rng, skip_first=True)(
        raw_observations,
    )

    assert len(gym_output.reward) == T - 1
    assert len(gym_info.agent.loss) == T - 1
    assert len(gym_info.agent.params["linear"]["w"]) == T - 1


class NonStationaryEnvironment(hk.Module):
    def __call__(self, action, raw_obs):
        step = hk.get_state("step", [], init=lambda *_: 0)

        # Only the environment now the true value of the parameters
        # at step 2000 we flip the sign of the true parameters !
        w_true = hk.cond(
            step < 2000,
            step,
            lambda step: -jnp.ones(3),
            step,
            lambda step: jnp.ones(3),
        )

        # The environment has its proper loss definition
        @jax.jit
        def loss(y_pred, y):
            return jnp.mean(jnp.square(y_pred - y))

        # raw observation contains features and generative noise
        x, noise = raw_obs

        # generate targets
        y = x @ w_true + noise
        obs = (x, y)

        # evaluate the prediction made by the agent
        y_previous = Lag(1)(y)
        y_pred = action
        reward = loss(y_pred, y_previous)

        step += 1
        hk.set_state("step", step)

        info = {}
        return reward, obs, info


def test_non_stationary_environement():
    # ## Non-stationary environment
    # Now, let's implement a non-stationary environment
    # Now let's run a gym simulation to see how the agent adapt to the change of environment.

    @hk.transform_with_state
    def gym_fun(raw_obs):
        return GymFeedback(
            linear_regression_agent, NonStationaryEnvironment(), return_action=True
        )(raw_obs)

    T = 300
    raw_observations = generate_many_raw_observations(T)
    rng = jax.random.PRNGKey(42)
    (gym_output, gym_info), final_state = unroll(
        gym_fun, return_final_state=True, skip_first=True, rng=rng
    )(raw_observations)

    assert len(gym_output.reward) == T - 1
    assert len(gym_info.agent.loss) == T - 1
    assert len(gym_info.agent.params["linear"]["w"]) == T - 1



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import haiku as hk
import jax
import jax.numpy as jnp
import pandas as pd
from jax.config import config

from wax.compile import jit_init_apply
from wax.modules.rolling_mean import RollingMean
from wax.unroll import unroll


def test_rolling_mean_init_apply():
    @jit_init_apply
    @hk.transform_with_state
    def rolling_mean(x):
        return RollingMean(2)(x)

    seq = hk.PRNGSequence(42)
    x = jax.random.normal(next(seq), (2, 3))
    params, state = rolling_mean.init(next(seq), x)
    output, state = rolling_mean.apply(params, state, next(seq), x)
    assert output is not x
    assert (output == x).all()
    x1 = x

    x = jax.random.normal(next(seq), (2, 3))
    output, state = rolling_mean.apply(params, state, next(seq), x)
    assert ((x1 + x) / 2 == output).all()
    x2 = x

    x = jax.random.normal(next(seq), (2, 3))
    output, state = rolling_mean.apply(params, state, next(seq), x)
    assert ((x2 + x) / 2 == output).all()


def test_run_ema_vs_pandas_not_adjust(window=10, min_periods=5):

    config.update("jax_enable_x64", True)

    seq = hk.PRNGSequence(42)
    x = jax.random.normal(shape=(100, 5), key=next(seq), dtype=jnp.float64)
    mask = jax.random.choice(next(seq), 2, shape=x.shape)
    x = jnp.where(mask, x, jnp.nan)

    @jit_init_apply
    @hk.transform_with_state
    def rolling_mean(x):
        return RollingMean(window, min_periods)(x)

    mean = unroll(rolling_mean)(x)
    mean_pandas = pd.DataFrame(x).rolling(window, min_periods).mean()

    assert jnp.allclose(mean, mean_pandas.values, equal_nan=True)



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import haiku as hk
import jax
import jax.numpy as jnp

from wax.modules.mask_mean import MaskMean


def test_mask_mean_no_mask():

    rng = hk.PRNGSequence(42)
    x = jax.random.normal(next(rng), (10,))
    x_mean_ref = x.mean()

    fun = hk.transform(lambda mask, x: MaskMean()(mask, x))
    mask = jnp.full(x.shape, True)

    params = fun.init(next(rng), mask, x)
    x_mean = fun.apply(params, next(rng), mask, x)
    assert jnp.allclose(x_mean, x_mean_ref)


def test_mask_mean_with_mask():

    rng = hk.PRNGSequence(42)
    x = jax.random.normal(next(rng), (10,))
    x_mean_ref = x[1:].mean()

    fun = hk.transform(lambda mask, x: MaskMean()(mask, x))

    mask = jnp.full(x.shape, True)
    mask = jax.ops.index_update(mask, jax.ops.index[:1], False)

    params = fun.init(next(rng), mask, x)
    x_mean = fun.apply(params, next(rng), mask, x)
    assert jnp.allclose(x_mean, x_mean_ref)


def test_mask_mean_with_mask_shape_2():
    rng = hk.PRNGSequence(42)
    x = jax.random.normal(next(rng), (2, 3))

    fun = hk.transform(lambda mask, x: MaskMean()(mask, x))

    mask = jnp.full(x.shape, True)
    mask = jax.ops.index_update(mask, jax.ops.index[0, 2], False)

    params = fun.init(next(rng), mask, x)
    x_mean = fun.apply(params, next(rng), mask, x)

    # mean_ref = jax.ops.index_update(x, ~mask, 0).mean()
    x_mean_ref = x[mask].mean()
    assert jnp.allclose(x_mean, x_mean_ref)


def test_mask_mean_axis_1():
    rng = hk.PRNGSequence(42)
    x = jax.random.normal(next(rng), (2, 3))

    fun = hk.transform(lambda mask, x: MaskMean(axis=0)(mask, x))

    mask = jnp.full(x.shape, True)
    mask = jax.ops.index_update(mask, jax.ops.index[0, 2], False)

    params = fun.init(next(rng), mask, x)
    x_mean = fun.apply(params, next(rng), mask, x)

    # the last axis has only one observation
    assert jnp.allclose(x_mean[2], x[1, 2])

    for col in range(3):
        x_mean_col_ref = x[:, col][mask[:, col]].mean()
        assert jnp.allclose(x_mean[col], x_mean_col_ref)
